"""
---------------------------------------------------------------------------------------------------------------------------------------
MÓDULO SMALLWORLDINESS
---------------------------------------------------------------------------------------------------------------------------------------
El módulo smallworliness implementa las principales medidas del nivel de aproximación de la estructura conectiva de un grafo a la
topología small world.

Estas vienen dadas por los coeficientes [1]
   1) SIGMA: tiene marcada tradición en el campo pero recientemente evidenciado como pésimo indicador por su fuerte dependencia con el
      número de nodos, grado q y densidad del grafo y por sus evidentes carencias conceptuales.

   2) OMEGA y OMEGA': miden la proximidad a un grafo Small World ideal en términos matemáticos (idealidad alcanzable). Son estables
      respecto a variaciones en el número de nodos, grado q y densidad del grafo. No obstante presentan ciertas carencias conceptuales
      lo que los hace susceptibles a dar falsos positivos.
      
   3) SWI: es una métrica bastante reciente y que mide la proximidad a un grafo Small World ideal en términos conceptuales (idealidad
      inalcanzable). Presenta cierta inestabilidad para número de nodos mayor o igual a 500 aunque es estable con el grado q. Su
      construcción lo hace susceptible a hacer interpretaciones excesivamente restrictivas (falsos negativos).

No obstante, las carencias de SWI solo se manifiestan al considerar los valores del coeficiente en términos absolutos (hacer una medida
de la que se pretenda obtener una idea de la small worldiness en términos absolutos). Si lo que se busca es una idea relativa de la
small worldiness de varios grafos este coeficiente es el adecuado dado que es conceptualmente más consistente que omega y omega' y menos
susceptible a dar falsos positivos.

REFERENCIAS
-----------

Durante la redacción se han hecho referencia a las siguientes publicaciones
  [1] Zachary P. Neal "How Small is it? Comparing indices of small worldliness" Network Science 5(1): 30-44, 2017

"""

import networkx          as nx
import random            as rd
import numpy             as np

from copy             import deepcopy
from networkx.classes import Graph, DiGraph

#--------------------------------------------------------------------------------------------------------------------------------------
# FUNCIONES GENERALES
#--------------------------------------------------------------------------------------------------------------------------------------

def distance_to_diagonal_matrix(N):

    """
    Función que calcula de la matriz contenedora de las distancias de cada elemento a la diagonal principal para una matriz cuadrada de
    grado N (matriz NxN).

    ARGUMENTOS
    ----------

    La función requiere de un argumento indispensable:
      ·N (int) : grado de la matriz cuadrada cuya matriz de distancias a la diagonal quiere computarse.

    SALIDA
    ------

    La función devuelve la matriz M en formato array de NumPy que contiene las distancias de cada elemento hasta la diagonal.

    NOTAS DE IMPLEMENTACIÓN
    -----------------------

    La implementación es un tanto 'pythónica' (alto nivel linguístico) y oscura pero es esto mismo lo que le aporta concisión y firmeza
    de forma que si se desea ahondar en los pormenores del algoritmo recomendamos consultar la documentación de la librería NumPy.

    """

    M  = np.zeros((N,N))
    v1 = np.arange(1,N)
    v2 = np.flip(v1)
    v  = np.append((0), np.where(v1<v2,v1,v2))

    for i in range(int(np.ceil(N/2))):
        M[N-i-1,:] = np.append(v[i+1:], v[:i+1])
        M[i,:]     = M[N-i-1,:][::-1]

    return(M)


def randomize(G, niter = 10):

    """
    Función randomizadora de grafos dirigidos y no dirigidos.

    ARGUMENTOS
    ----------

    La función requiere de dos argumentos, uno de ellos obligatorio y el otro opcional:
      · G     (G/DiG) : instancia de la clase G/DiG definida en el entorno NetworkX
      · niter (int)   : número aproximado de veces que una arista resultará recableada

    SALIDA
    ------

    La función devuelve un grafo G (NetworkX Graph/DiGraph) resultado del proceso de randomizado del grafo introducido como argumento.

    NOTAS DE IMPLEMETNACIÓN
    -----------------------

    El algoritmo de randomizado ha sido adaptado del módulo smallworld.py de NetworkX que, a su vez adapta el algoritmo propuesto
    por Maslov y Sneppen (2002) [1].

    En esta adaptación se incluyen  diferencias notables respecto a los algoritmos de Maslov-Sneppen y el adaptado en NetworkX, como la
    generalización del algoritmo a grafos dirigidos, la predilección no solo por nodos hubs (con alto grado de salida) sino también por
    nodos authorities (con alto grado de entrada) a la hora de materializar el recableado.

    El algoritmo empleado es realmente instructor, paradigmático y estético, si dispone de tiempo se recomienda al lector que ahonde en
    sus pormenores.

    REFERENCIAS
    -----------

    Durante la redacción se han hecho referencia a las siguientes publicaciones
      [1] Maslov, S. ; Sneppen, K. "Specificity and stability in topology of protein networks." Science 296.5569 (2002): 910-913.

    """

    from networkx.utils import cumulative_distribution, discrete_sequence

    G        = deepcopy(G)
    directed = G.is_directed()

    if not directed:
        keys, degrees = zip(*G.degree())
        cdf           = nx.utils.cumulative_distribution(degrees)
        nnodes        = len(G)
        nedges        = nx.number_of_edges(G)

        niter         = niter*nedges
        ntries        = (nnodes*nedges/(nnodes*(nnodes-1)/2))
        swapcount     = 0

        for i in range(niter):
            n = 0
            while n < ntries:
                ai,ci = discrete_sequence(2, cdistribution=cdf)
                if ai == ci:
                    continue
                a,c = keys[ai], keys[ci]  # de índice a etiqueta nodal
                b,d = [rd.choice(list(G.neighbors(x))) for x in [a,c]]

                if b in [a,c,d] or d in [a, b, c]:
                    continue              # todos los nodos deben ser diferentes

                if (d not in G[a]) and (b not in G[c]):
                    G.add_edge(a,d)
                    G.add_edge(c,b)
                    G.remove_edge(a,b)
                    G.remove_edge(c,d)

                    swapcount += 1
                else:
                    n += 1

    if directed:
        keys_out, out_degrees = zip(*G.out_degree())

        cdf_out       = nx.utils.cumulative_distribution(out_degrees)
        nnodes        = len(G)
        nedges        = nx.number_of_edges(G)

        niter         = niter*nedges
        ntries        = (nnodes*nedges/(nnodes*(nnodes-1)/2))
        swapcount     = 0

        for i in range(niter):
            n = 0
            while n < ntries:
                ai,ci = discrete_sequence(2, cdistribution=cdf_out) # predilección por hubs
                if ai == ci:
                    continue
                a, c = keys_out[ai], keys_out[ci]

                succ_a, succ_c = list(G.successors(a)), list(G.successors(c))

                keys_in_a , in_degrees_a  = zip(*G.in_degree(succ_a))
                keys_in_c , in_degrees_c  = zip(*G.in_degree(succ_c))
                cdf_in_a = nx.utils.cumulative_distribution(in_degrees_a)
                cdf_in_c = nx.utils.cumulative_distribution(in_degrees_c)

                bi = discrete_sequence(1, cdistribution = cdf_in_a)[0]  # predilección por authorities
                di = discrete_sequence(1, cdistribution = cdf_in_c)[0]  # predilección por authorities
                b  = keys_in_a[bi]
                d  = keys_in_c[di]

                if b in [a, c, d] or d in [a, b, c]:
                    continue

                if (d not in G.successors(a)) and (b not in G.successors(c)):
                    G.add_edge(a,d)
                    G.add_edge(c,b)
                    G.remove_edge(a,b)
                    G.remove_edge(c,d)

                    swapcount += 1
                else:
                    n += 1

    return G


def latticize(G,niter = 10):

    """
    Función latticizadora de grafos dirigidos y no dirigidos.

    ARGUMENTOS
    ----------

    La función requiere de dos argumentos, uno de ellos obligatorio y el otro opcional:
      · G     (G/DiG) : instancia de la clase G/DiG definida en el entorno NetworkX
      · niter (int)   : número aproximado de veces que una arista resultará recableada

    SALIDA
    ------

    La función devuelve un grafo G (NetworkX Graph/DiGraph) resultado del proceso de latticizado del grafo introducido como argumento.

    NOTAS DE IMPLEMETNACIÓN
    -----------------------

    El algoritmo de latticizado ha sido adaptado del módulo smallworld.py de NetworkX que, a su vez, está adaptada del algoritmo
    propuesto por Sporns et al. [1] inspirado en el trabajo original de Maslov y Sneppen (2002) [2].

    En esta adaptación se incluyen  diferencias notables respecto a los algoritmos de Maslov-Sneppen y el adaptado en NetworkX, como la
    generalización del algoritmo a grafos dirigidos, la predilección no solo por nodos hubs (con alto grado de salida) sino también por
    nodos authorities (con alto grado de entrada) a la hora de materializar el recableado.

    La implementación es exactamente igual a la de la función randomize salvo por la adición de una condición extra que solo permite
    la materialización del recableado de aristas (X-swap) si esto induce una aproximación sucesiva de la matriz de adyacencia del grafo
    a una matriz diagonal representativa de una topología tipo lattice (red regular).

    El algoritmo empleado es realmente instructor, paradigmático y estético, si dispone de tiempo se recomienda al lector que ahonde en
    sus pormenores.

    REFERENCIAS
    -----------

    Durante la redacción se han hecho referencia a las siguientes publicaciones
      [1] Sporns, Olaf, and Jonathan D. Zwi. "The small sorld of the cerebral cortex." Neuroinfrmatics 2.2 (2004): 145-162
      [2] Maslov, S. ; Sneppen, K. "Specificity and stability in topology of protein networks." Science 296.5569 (2002): 910-913.

    """

    import numpy as np
    from networkx.utils import cumulative_distribution, discrete_sequence

    G = deepcopy(G)
    directed = G.is_directed()

    if not directed:
        keys, degrees = zip(*G.degree())
        cdf           = nx.utils.cumulative_distribution(degrees)
        nnodes        = len(G)
        nedges        = nx.number_of_edges(G)

        niter         = niter*nedges
        ntries        = (nnodes*nedges/(nnodes*(nnodes-1)/2))
        swapcount     = 0

        D = distance_to_diagonal_matrix(nnodes)

        for i in range(niter):
            n = 0
            while n < ntries:
                ai,ci = discrete_sequence(2, cdistribution=cdf)
                if ai == ci:
                    continue
                a,c = keys[ai], keys[ci]
                b   = rd.choice(list(G.neighbors(a)))
                d   = rd.choice(list(G.neighbors(c)))
                bi  = keys.index(b)
                di  = keys.index(d)

                if b in [a,c,d] or d in [a, b, c]:
                    continue

                if (d not in G[a]) and (b not in G[c]):
                    if D[a,b] + D[c,d] >= D[a,c] + D[b,d]:
                        G.add_edge(a,d)
                        G.add_edge(c,b)
                        G.remove_edge(a,b)
                        G.remove_edge(c,d)

                        swapcount += 1
                else:
                    n += 1
    if directed:
        keys_out, out_degrees = zip(*G.out_degree())

        cdf_out       = nx.utils.cumulative_distribution(out_degrees)
        nnodes        = len(G)
        nedges        = nx.number_of_edges(G)

        niter         = niter*nedges
        ntries        = (nnodes*nedges/(nnodes*(nnodes-1)/2))
        swapcount     = 0

        D = distance_to_diagonal_matrix(nnodes)

        for i in range(niter):
            n = 0
            while n < ntries:
                ai,ci = discrete_sequence(2, cdistribution=cdf_out) # predilección por hubs
                if ai == ci:
                    continue
                a, c = keys_out[ai], keys_out[ci]

                succ_a, succ_c = list(G.successors(a)), list(G.successors(c))

                keys_in_a , in_degrees_a  = zip(*G.in_degree(succ_a))
                keys_in_c , in_degrees_c  = zip(*G.in_degree(succ_c))
                cdf_in_a = nx.utils.cumulative_distribution(in_degrees_a)
                cdf_in_c = nx.utils.cumulative_distribution(in_degrees_c)

                bi = discrete_sequence(1, cdistribution = cdf_in_a)[0]  # predilección por authorities
                di = discrete_sequence(1, cdistribution = cdf_in_c)[0]  # predilección por authorities
                b  = keys_in_a[bi]
                d  = keys_in_c[di]

                if b in [a, c, d] or d in [a, b, c]:
                    continue

                if (d not in G.successors(a)) and (b not in G.successors(c)):
                    if D[a,b] + D[c,d] >= D[a,c] + D[b,d]:
                        G.add_edge(a,d)
                        G.add_edge(c,b)
                        G.remove_edge(a,b)
                        G.remove_edge(c,d)

                        swapcount += 1
                else:
                    n += 1

    return G

#--------------------------------------------------------------------------------------------------------------------------------------
# FUNCIONES ESPECÍFICAS
#--------------------------------------------------------------------------------------------------------------------------------------

def sigma (G, niter = 100, nrand = 10, RG = None, sup = False):

    """
    Devuelve el coeficiente de Small Worldness sigma definido como

                                                    sigma = (C/Cran) / (L/Lran)                                                 [1]

    donde C y L son respectivamente el average clustering coefficient y el average shortest path length del grafo, las subetiquetas
    "ran" hacen referencia al grafo randomizado asociado.

    Una revisión reciente dede Zachary [1] pone pone en evidencia que el coeficiente sigma, aunque ha sido el más ampliamente adoptado
    en la literatura hasta el momento, no es, ni mucho menos, el mejor indicador de Small Worldiness dado que su simplicidad corre a
    costa de
      - una evidente inconsistencia con respecto a la concepción de Watts-Strogatz de la estructura Small World
      - una fuerte dependencia sobre los parámetros N y k que limitan su aplicabilidad

    Así pues, únicamente se incluye la implementación de el coeficiente debido a su amplia extensión y como herramienta que permita al
    usuario comprobar sus limitaciones y carencias.

    ARGUMENTOS
    ----------

    La función sigma requiere de un argumento obligatorio y cuatro opcionales
      · G     (G/DiG): instancia de la clase G/DiG definida en el entorno NetworkX
      · niter (int)  : número aproximado de recableados por arista para obtener el grafo randomizado asociado
      · nrand (int)  : número de grafos randomizados generados para obtener los coeficientes promedio Cran y Lran
      · RG    (list) : lista de grafos randomizados que opcionalmente pueden proveerse para agilizar el proceso.
      · sup   (bool) : cuando True devuelve además de S un diccionario con los valores Cran y Lran de los nran grafos randomizados.

    No existen restricciones sobre el tamaño de RG y su objetivo es agilizar el proceso cuando se quieren calcular varios coeficientes
    para el mismo grafo WS (p.ej. si se quieren realizar comparaciones del comportamiento de coeficientes). Así, pueden crearse listas
    de nrand grafos randomizados y nlat grafos lattizados asociados a una determinada instanciación de WS y  pasarlas a los métodos
    sigma, omega y SWI para que cada uno de ellos no tenga que crear grafos andomizados y/o lattizados, agilizando así el proceso.

    SALIDA
    ------

    La salida depende del argumento sup de forma que
      - Si sup = False: la función devuelve únicamente el coeficiente sigma del grafo considerado.
      - Si sup = True : la función devuelve tanto el coeficiente sigma del grafo considerado como un diccionario con los valores Cran y
                        Lran de los nran grafos randomizados.

    NOTAS DE IMPLEMENTACIÓN
    -----------------------

    Los valores representativos de Lran y Cran se obtienen como media de los respectivos valores para nran grafos randomizados a partir
    del original.

    REFERENCIAS
    -----------

    Durante la redacción se ha hecho referencia a las siguientes publicaciones
    [1] M.D. Humphries, K. Gurney and T.J. Prescott "The brainstem reticular formation in a small-world, not scale-free, network."
        Proc. R. Soc. B (2006) 273, 503-511

    """

    if RG != None:
        G_ran_list = RG
        nrand      = len(G_ran_list)
    else:
        G_ran_list = [randomize(G,niter = niter) for i in range(nrand)]


    coeficientes = {"Cran": [], "Lran": []}

    for i in range(nrand):
        G_ran = G_ran_list[i]
        coeficientes["Cran"].append(nx.average_clustering(G_ran))
        coeficientes["Lran"].append(nx.average_shortest_path_length(G_ran))

    C = nx.average_clustering(G)
    L = nx.average_shortest_path_length(G)

    Cran_av = np.mean(coeficientes["Cran"])
    Lran_av = np.mean(coeficientes["Lran"])

    sigma = (C/Cran_av)/(L/Lran_av)

    if sup:
        return sigma, coeficientes
    else:
        return sigma


def omega (G, niter = 100, nrand = 10, nlat = 10, RG = None, LG = None, sup = False):

    """
    Devuelve el coeficiente de Small Worldness omega definido por como
                                                      omega = Lran/L - C/Clat                                                    [1]
    donde C y L son respectivamente el average clustering coefficient y el average shortest path length de G. Las subetiquetas lat y
    ran hacen eferencia a los grafos lattizado y randomizado asociados al grafo original.

    El coeficiente omega es interpretado en un espectro continuo en el que:
      (1) Omega = -1 --> Lattice Graph
      (2) Omega =  0 --> Small World Graph
      (3) Omega = +1 --> Random Graph

    El coeficiente omega mide la proximidad a un grafo Small World ideal en términos matemáticos (idealidad alcanzable). Es estables
    respecto a variaciones en el número de nodos, grado q y densidad del grafo. No obstante presenta ciertas carencias conceptuales
    resultando propenso a dar falsos positivos [1].

    ARGUMENTOS
    ----------

    La función omega requiere de un argumento obligatorio y seis opcionales:
      · niter (int) : número aproximado de recableados por arista para obtener el grafo randomizado/latticizado asociado
      · nrand (int) : número de grafos randomizados generados para obtener los coeficientes promedio Clat y Lran
      · RG    (list): lista de grafos randomizados
      · LG    (list): lista de grafos latizados
      · sup   (bool): cuando True devuelve además de O un diccionario con los valores de Lran de los nran grafos randomizados y Clat de los
                      nlat grafos latticizados.

    No existen restricciones sobre los tamaños de RG y LG y su objetivo es agilizar el proceso cuando se quieren calcular diferentes
    coeficientes para el mismo grafo WS (por ejemplo si se quieren realizar comparaciones del comportamiento de coeficientes para un
    mismo grafo). Así, pueden crearse listas de nrand grafos randomizados y nlat grafos lattizados asociados a una instanciación de
    SmallWorld y pasarlas a los métodos sigma, omega y SWI para que cada uno de ellos no tenga que crear grafos randomizados y/o
    lattizados, agilizando así el proceso.

    SALIDA
    ------

    La salida depende del argumento sup de forma que
      - Si sup = False: la función devuelve únicamente el coeficiente omega del grafo considerado.
      - Si sup = True : la función devuelve tanto el coeficiente omega del grafo considerado como un diccionario con los valores Cran y
                         Lran de los nran grafos randomizados.

    NOTAS DE IMPLEMENTACIÓN
    ------------------------

    Los valores representativos de Lran y Clat se obtienen como media de los respectivos valores para nrand grafos randomizados y nlat
    grafos lattizados a partir del original.

    REFERENCIAS
    -----------

    Durante la redacción se ha hecho referencia a las siguientes publicaciones:
      [1] Telesfold, Q.K et.al. "The Ubiquity of Small World Networks" Brain Connectivity 1.5 (2011)

    """

    if RG != None and LG != None:
        G_ran_list = RG
        nrand      = len(G_ran_list)

        G_lat_list = LG
        nlat       = len(G_lat_list)

    elif RG != None and LG == None:
        G_ran_list = RG
        nrand      = len(G_ran_list)

        G_lat_list = [latticize(G,niter = niter) for i in range(nlat)]

    elif RG == None and LG != None:
        G_ran_list = [randomize(G,niter = niter) for i in range(nrand)]

        G_lat_list = LG
        nlat       = len(G_lat_list)

    elif RG == None and LG == None:
        G_ran_list = [randomize(G,niter = niter) for i in range(nrand)]
        G_lat_list = [latticize(G,niter = niter) for i in range(nlat)]


    coeficientes = {"Clat": [], "Lran": []}

    for i in range(nrand):
        G_ran = G_ran_list[i]
        coeficientes["Lran"].append(nx.average_shortest_path_length(G_ran))

    for i in range(nlat):
        G_lat = G_lat_list[i]
        coeficientes["Clat"].append(nx.average_clustering(G_lat))

    C = nx.average_clustering(G)
    L = nx.average_shortest_path_length(G)

    Clat_av = np.mean(coeficientes["Clat"])
    Lran_av = np.mean(coeficientes["Lran"])

    omega = (Lran_av/L) - (C/Clat_av)

    if sup:
        return omega, coeficientes
    else:
        return omega


def omega_p (G, niter = 100, nrand = 10, nlat = 10, RG = None, LG = None, sup = False):

    """
    Devuelve el coeficiente de Small Worldness omega' definido por como
                                                      omega' = 1 - abs(omega)                                                    [1]
    donde C y L son respectivamente el average clustering coefficient y el average shortest path length de G. Las subetiquetas lat y
    ran hacen eferencia a los grafos lattizado y randomizado asociados al grafo original.

    El coeficiente omega es interpretado en un espectro continuo en el que:
      (1) Omega' = 0 --> Lattice or Random graph
      (2) Omega' = 1 --> Small World Graph

    El coeficiente omega' mide la proximidad a un grafo Small World ideal en términos matemáticos (idealidad alcanzable). Es estables
    respecto a variaciones en el número de nodos, grado q y densidad del grafo. No obstante presenta ciertas carencias conceptuales
    resultando propenso a dar falsos positivos [1].

    ARGUMENTOS
    ----------

    La función omega requiere de un argumento obligatorio y seis opcionales:
      · niter (int) : número aproximado de recableados por arista para obtener el grafo randomizado/latticizado asociado
      · nrand (int) : número de grafos randomizados generados para obtener los coeficientes promedio Clat y Lran
      · RG    (list): lista de grafos randomizados
      · LG    (list): lista de grafos latizados
      · sup   (bool): cuando True devuelve además de O un diccionario con los valores de Lran de los nran grafos randomizados y Clat de los
                      nlat grafos latticizados.

    No existen restricciones sobre los tamaños de RG y LG y su objetivo es agilizar el proceso cuando se quieren calcular diferentes
    coeficientes para el mismo grafo WS (por ejemplo si se quieren realizar comparaciones del comportamiento de coeficientes para un
    mismo grafo). Así, pueden crearse listas de nrand grafos randomizados y nlat grafos lattizados asociados a una instanciación de
    SmallWorld y pasarlas a los métodos sigma, omega y SWI para que cada uno de ellos no tenga que crear grafos randomizados y/o
    lattizados, agilizando así el proceso.

    SALIDA
    ------

    La salida depende del argumento sup de forma que
      - Si sup = False: la función devuelve únicamente el coeficiente omega' del grafo considerado.
      - Si sup = True : la función devuelve tanto el coeficiente omega' del grafo considerado como un diccionario con los valores Cran y
                         Lran de los nran grafos randomizados.

    NOTAS DE IMPLEMENTACIÓN
    ------------------------

    Los valores representativos de Lran y Clat se obtienen como media de los respectivos valores para nrand grafos randomizados y nlat
    grafos lattizados a partir del original.

    REFERENCIAS
    -----------

    Durante la redacción se ha hecho referencia a
      [1] Telesfold, Q.K et.al. "The Ubiquity of Small World Networks" Brain Connectivity 1.5 (2011)

    """

    if sup:
        omega_c, coeficientes = omega(G, niter = niter, nrand = nrand, nlat = nlat, RG = RG, LG = LG, sup = sup)
        omega_p = 1 - abs(omega_c)
        return omega_p, coeficientes
    else:
        omega_p = 1 - abs(omega(G,niter = niter, nrand = nrand, nlat = nlat, RG = RG, LG = LG, sup = sup))
        return omega_p


def SWI (G, niter = 100, nrand = 10, nlat = 10, RG = None, LG = None, sup = False):

    """
    Devuelve el coeficiente de Small Worldness omega' definido por como
                                    SWI = [ (L-Llat) / (Lran-Llat) ] * [ (C-Cran) / (Clat-Cran) ]                                [1]
    donde C y L son respectivamente el average clustering coefficient y el average shortest path length de G. Las subetiquetas lat y
    ran hacen referencia a los grafos lattizado y randomizado asociados al grafo original.

    El coeficiente omega es interpretado en un espectro continuo en el que:
      (1) SWI = 0 --> El grafo solo presenta una de las dos propiedades características (L pequeño y C alto) de las topologías Small
                      World o ninguna de ellas.
      (2) SWI = 1 --> El grafo presenta las dos características de las topologías Small World.

    El coeficiente SWI s una métrica bastante reciente y balanceada, que mide la proximidad a un grafo Small World ideal en términos
    conceptuales (idealidad inalcanzable). Presenta cierta inestabilidad para número de nodos mayor o igual a 500 aunque es estable con
    el grado q. Su construcción lo hace susceptible a hacer interpretaciones excesivamente restrictivas (falsos negativos) [1].

    ARGUMENTOS
    ----------

    La función omega requiere de un argumento obligatorio y seis opcionales:
      · niter (int) : número aproximado de recableados por arista para obtener el grafo randomizado/latticizado asociado
      · nrand (int) : número de grafos randomizados generados para obtener los coeficientes promedio Clat y Lran
      · RG    (list): lista de grafos randomizados
      · LG    (list): lista de grafos latizados
      · sup   (bool): cuando True devuelve además de O un diccionario con los valores de Lran de los nran grafos randomizados y Clat de los
                      nlat grafos latticizados.

    No existen restricciones sobre los tamaños de RG y LG y su objetivo es agilizar el proceso cuando se quieren calcular diferentes
    coeficientes para el mismo grafo WS (por ejemplo si se quieren realizar comparaciones del comportamiento de coeficientes para un
    mismo grafo). Así, pueden crearse listas de nrand grafos randomizados y nlat grafos lattizados asociados a una instanciación de
    SmallWorld y pasarlas a los métodos sigma, omega y SWI para que cada uno de ellos no tenga que crear grafos randomizados y/o
    lattizados, agilizando así el proceso.

    SALIDA
    ------

    La salida depende del argumento sup de forma que
      - Si sup = False: la función devuelve únicamente el coeficiente SWI del grafo considerado.
      - Si sup = True : la función devuelve tanto el coeficiente SWI del grafo considerado como un diccionario con los valores Cran y
                         Lran de los nran grafos randomizados.

    NOTAS DE IMPLEMENTACIÓN
    ------------------------

    Los valores representativos de Lran y Clat se obtienen como media de los respectivos valores para nrand grafos randomizados y nlat
    grafos lattizados a partir del original.

    REFERENCIAS
    -----------
    Durante la redacción se ha hecho referencia a las siguientes publicaciones:
      [1] Zachary P.N. "How small is it? Comparing indices of small worldliness." Network Science 5(1) (2017): 30-44

    """

    import numpy as np

    if RG != None and LG != None:
        G_ran_list = RG
        nrand      = len(G_ran_list)

        G_lat_list = LG
        nlat       = len(G_lat_list)

    elif RG != None and LG == None:
        G_ran_list = RG
        nrand      = len(G_ran_list)

        G_lat_list = [latticize(G,niter = niter) for i in range(nlat)]

    elif RG == None and LG != None:
        G_ran_list = [randomize(G,niter = niter) for i in range(nrand)]

        G_lat_list = LG
        nlat       = len(G_lat_list)

    elif RG == None and LG == None:
        G_ran_list = [randomize(G,niter = niter) for i in range(nrand)]
        G_lat_list = [latticize(G,niter = niter) for i in range(nlat)]


    coeficientes = {"Clat": [], "Llat": [], "Cran": [], "Lran": []}

    for i in range(nrand):
        G_ran = G_ran_list[i]
        coeficientes["Lran"].append(nx.average_shortest_path_length(G_ran))
        coeficientes["Cran"].append(nx.average_clustering(G_ran))

    for i in range(nlat):
        G_lat = G_lat_list[i]
        coeficientes["Llat"].append(nx.average_shortest_path_length(G_lat))
        coeficientes["Clat"].append(nx.average_clustering(G_lat))

    C = nx.average_clustering(G)
    L = nx.average_shortest_path_length(G)

    Clat_av = np.mean(coeficientes["Clat"])
    Cran_av = np.mean(coeficientes["Cran"])
    Llat_av = np.mean(coeficientes["Llat"])
    Lran_av = np.mean(coeficientes["Lran"])

    SWI = ((L-Llat_av)/(Lran_av-Llat_av)) * ((C-Cran_av)/(Clat_av-Cran_av))

    if sup:
        return SWI, coeficientes
    else:
        return SWI


#--------------------------------------------------------------------------------------------------------------------------------------
# NOTAS AL PIE
#--------------------------------------------------------------------------------------------------------------------------------------

# (*) DESPUÉS DE MUCHOS PROBLEMAS: deepcopy vs shallowcopy (profunda vs superficial). Las copys superficiales (las usuales) se
#     materializan mediante referencias, no mediante copias a nuevas entidades, por tanto si el objeto a copiar es mutable o contiene
#     elementos mutables esto puede suponer problemas o efectos indeseados.
