"""
---------------------------------------------------------------------------------------------------------------------------------------
MÓDULO WATTSSTROGATZ
---------------------------------------------------------------------------------------------------------------------------------------

El módulo wattsstrogatz es un módulo de apoyo y referencia que contiene la implementación más básica del algoritmo de WS para un grafo
no dirigido. El módulo sienta sus bases en el paquete NetworkX del que hereda la metaclase Graph. 

"""

import networkx          as nx
import random            as rd

from networkx.classes import Graph


#-------------------------------------------------------------------------------
# CLASE WATTS STROGATZ
#-------------------------------------------------------------------------------

class WattsStrogatz(Graph):
    """

    ATRIBUTOS
    ---------
    Subclase de la clase Graph de Networkx's que implementa un tipo concreto de grafos no dirigidos: grafos construidos por el algoritmo
    de Watts-Strogatz.
    """

    def __init__(self, N, q, p, incoming_graph_data = None, **attr):

        Graph.__init__(self, incoming_graph_data, **attr)

        self.number_nodes         = N
        self.initial_degree       = q
        self.rewiring_probability = p

        self.add_nodes_from(list(range(self.number_nodes)))
        self._watts_strogatz_graph()

        self.graph["info"] = "Watts Strogatz graph with N = {0} nodes, q = {1} nearest neighbors and p = {2} rewiring probability".format(self.number_nodes, self.initial_degree, self.rewiring_probability)


    def _watts_strogatz_graph(self):

        """
        Crea un N-grafo con topología de Watts-Strogatz siguiendo el algoritmo WSWA (Watts Strogatz Small World Algorithm) con probab.
        de recableado p y grado de salida hacia vecinos proximos q.

        COMENTARIOS
        -----------
        El algoritmo parte de la creación de un N-grafo con topología anular y grado de salida uniforme hacia q vecinos próximos para
        seguidamente inducir un procedimiento de recableado sobre cada vértice, con probab. de materialización p, de acuerdo con la
        propuesta de Watts-Strogatz [1].

        REFERENCIAS
        -----------
        [1] Watts, D.J. ; Strogatz, S.H.
            "Collective dynamics of 'Small-World' networks"
            Nature 393.6684 (2002): 440-2
        """
  # (1) Creación de una topología anular

        N, q, p = len(self), self.initial_degree, self.rewiring_probability
        nodos = list(self.nodes())

        for i in range(1, q //2 + 1):
            receptores = nodos[i:] + nodos[:i]
            self.add_edges_from(zip(nodos,receptores))

  # (2) Implementación del algoritmo de recableado de Watts-Strogatz

        for edge in self.edges:
            n1,n2 = edge
            if rd.random() < p:
                n_new = rd.choice(nodos)
                while n_new == n1 or ((n1,n_new) in self.edges):
                    n_new = rd.choice(nodos)

                    if self.degree(n1) >= N-1:
                        break
                else:
                    self.remove_edge(n1,n2)
                    self.add_edge(n1,n_new)


#--------------------------------------------------------------------------
# PRUEBAS
#--------------------------------------------------------------------------
if __name__ == "__main__":

    from smallworldiness import *

    WS    = WattsStrogatz(N = 100, q = 10, p = 0.3)

    print(nx.is_connected(WS))
    print(nx.average_shortest_path_length(WS))

    n_niter = 1
    n_nrand = 1
    n_nlat  = 1

    randomized = [randomize(WS, niter = n_niter) for i in range(n_nrand)]
    latticized = [latticize(WS, niter = n_niter) for i in range(n_nlat )]

    print("Grafos auxiliares creados")

    print(sigma(WS, RG = randomized))
    print(omega_p(WS, RG = randomized, LG = latticized))
    print(SWI(WS, RG = randomized, LG = latticized))

    nx.write_graphml(WS, "undirected100.graphml")
