"""
---------------------------------------------------------------------------------------------------------------------------------------
MÓDULO CNS
---------------------------------------------------------------------------------------------------------------------------------------

El módulo CNS contiene la metaestructura OOP necesaria para implementar subestructuras anatómico-funcionales del SNC humano y consta de
4 niveles de profundidad jerárquica de acuerdo con la disposición anatómico-funcional convencional del SNC humano (no se considera la
médula espinal).

En concreto, el módulo integra los modelos de dinámica unineuronal implementados en el módulo neuron y los modelos de arquitectura
conectiva plasmados en el módulo connectome y desarrolla un modelo de neocórtex humano construido a partir de una arquitectura conectiva
tipo Small-World y unidades neuronales de acuerdo al modelo LIF (Leaky Integrate and Fire).

La estructura ha sido diseñada para facilitar incorporaciones y adiciones futuras de nuevos modelos y elementos.

"""

import numpy             as np
import scipy             as sp
import random            as rd
import matplotlib.pyplot as plt
import matplotlib        as mplt
import seaborn           as sns
import os                as os
import pandas            as pd
import networkx          as nx

import pickle

from collections     import defaultdict
from connectome      import DiWattsStrogatz
from neuron          import LIFN, Piramidal_NC, Interneuron_NC
from pandas          import DataFrame, Series

from copy            import deepcopy
from datetime        import datetime

#--------------------------------------------------------------------------------------------------------------------------------------
# GENERAL LAMBDAS
#--------------------------------------------------------------------------------------------------------------------------------------

recursive_dictionary = lambda : defaultdict(recursive_dictionary)

#--------------------------------------------------------------------------------------------------------------------------------------
# GENERAL FUNCTIONS
#--------------------------------------------------------------------------------------------------------------------------------------

def default_to_regular(d):

    """
    Conversor de arbol de default_dict anidados a usual dict, si d no es un defaultdict devuelve d inmutado.

    """
    if isinstance(d, defaultdict):
        d = {k: default_to_regular(v) for k, v in d.items()}
    return d

#--------------------------------------------------------------------------------------------------------------------------------------
# METACLASE ESTRUCTURAL DE PRIMER ORDEN (PRIMERA GENERACIÓN PARENTAL)
#--------------------------------------------------------------------------------------------------------------------------------------

class HCNS():

    """
    Human Central Nervous System (HCNS). Metaclase de primera generación (orden 1) para la modelización de estructuras neuronales propias
    de la anatomía humana y cuya única pretensión es vertebrar el esquema de programación orientada a objetos escogido.

    El SNC humano consta de 3 subestructuras funcionales principales (no se considera la médula espinal): el cerebro, el cerebelo y el
    tronco encefálico. En contra de lo que se creía hasta el momento los últimos estudios [1] indican que el SNC humano consta de un
    número similar de células neuronales (86.1 mil millones) y gliales (84.6 mil millones).

    REFERENCIAS
    ------------

    Durante la redacción se han hecho referencia a las siguientes publicaciones
       [1] C. Ragsdale, F1000 - Post-Publication Peer Review of the Biomedical Literature (2009).

    """

    pass

#--------------------------------------------------------------------------------------------------------------------------------------
# METACLASE ESTRUCTURAL DE SEGUNDO ORDEN (SEGUNDA GENERACIÓN PARENTAL)
#--------------------------------------------------------------------------------------------------------------------------------------

class Cerebrum(HCNS):

    """
    Metaclase de segunda generación (orden 2) para la modelización del cerebro humano.

    En condiciones fisiológicas el cerebro representa el 81.9 % de la masa y 19 % del número de neuronas del SNC humano y puede
    subdividirse en estructuras corticales (cortex cerebral) y estructuras subcorticales (hipocampo, amigdala, bulbo olfatorio ...).

    Se trata de la subestructura del SNC más reciente en términos evolutivos y es la encargada de las funciones cerebrales superiores,
    es decir, aquellas funciones que diferencian más marcadamente al ser humano del resto de animales, como pueden ser la conciencia,
    el pensamiento, la memoria, el razonamiento, las emociones, el aprendizaje, el lenguaje, ...

    """

    pass


class Cerebelum(HCNS):

    """
    Metaclase de segunda generación (orden 2) para la modelización del cerebelo humano.

    En condiciones fisiológicas el cerebro representa el 10.3 % de la masa y 80.2 % del número de neuronas del SNC humano.

    Se trata de la subestructura del SNC encargada de la coordinación funcional y el control de movimientos voluntarios como la
    postura, el equilibrio, el habla ...

    """

    pass


class Brainstem(HCNS):

    """
    Metaclase de segunda generación (orden 2) para la modelización del cerebro humano.

    En condiciones fisiológicas el cerebro representa el 7.8 % de la masa y 0.8 % del número de neuronas del SNC humano.

    Se trata de la subestructura del SNC encargada de la coordinación funcional y el control de movimientos involuntarios como la
    respiración, el ritmo cardíaco, la presión sanguínea, ...

    """

    pass

#--------------------------------------------------------------------------------------------------------------------------------------
# METACLASE ESTRUCTURAL DE TERCER ORDEN (TERCERA GENERACIÓN PARENTAL)
#--------------------------------------------------------------------------------------------------------------------------------------

class Cortex(Cerebrum):

    """
    Metaclase de tercera generación (orden 3) para la modelización del cortex cerebral humano.

    El tejido cortical del cerebro humano conforma el 98 % del cerebro humano (80 % del SNC) y consta de 9 capas, 6 de las cuales
    corresponden a tejido nocortical (neocortex), siendo las 3 restantes capas allocorticales. Por tanto, en el caso humano el cortex
    cerebral resulta ser esencialmente neocortex.
    """

    pass

#--------------------------------------------------------------------------------------------------------------------------------------
# METACLASE ESTRUCTURAL DE CUARTO ORDEN (CUARTA GENERACIÓN PARENTAL)
#--------------------------------------------------------------------------------------------------------------------------------------

class Neocortex(Cortex):

    """
    Metaclase de cuarta generación (orden 4) para la modelización del neocortex cerebral humano.

    El neocortex humano es el culmen de la evolución del SNC y el epicentro de las funciones cerebrales superiores, especialmente de
    la percepción, la toma de decisiones y el lenguaje. Además, conforma prácticamente el 80 % del SNC humano por lo que su estudio
    resulta de gran relevancia.

    En condiciones fisiológicas contiene un
       - 80% de neuronas proyectivas (principalmente piramidales) que son excitatorias (glutamatergicas)
       - 20% de neuronas intrínsecas (agrupadas bajo el término interneurona) que son inhibitorias (GABAergicas)

    ATRIBUTOS DE CLASE
    ------------------

    Los atributos de clase asociados a la metaclase Neocortex son los siguientes:
      · base_path (str) : directorio base para guardar datos de simulaciones, estructuras etc.

    ATRIBUTOS DE INSTANCIA
    ----------------------

    Los atributos de instancia asociados a la metaclase Neocortex son los siguientes:
      · neurons_dic     (dic)     : diccionario que contiene todas las neuronas que tomarán parte en la construcción del Neocortex
                                    indexadas por los instanciadores de los tipos neuronales que se consideren.
      · digraph_info    (str)     : información a cerca de la estructura conectiva subyacente (grafo dirigido) y que está contenida en
                                    el objeto portador de dicha estructura
      · model_info      (str)     : información a cerca del modelo de dinámica unineuronal empleado y que está contenida en cualquiera
                                    de los objetos portadores de dicha dinámica.
      · digraph         (DiGraph) : objeto DiGraph de la librería NetworkX que contiene la estructura conectiva del neocortex modelizado
                                    en forma de grafo dirigido.
      · neocortex       (DiGraph) : objeto DiGraph consecuencia de la proyección de la diámica unineuronal a través de la estructura
                                    conectiva subyacente.
      · dt              (float)   : parámetro de discretización temporal de cara a las simulaciones
      · simulation_time (float)   : tiempo durante el que desea monitorizarse la simulación (tiempo propio del sistema simulado)
      · simulation_long (float)   : tiempo externo que toma el tratamiento computacional de la simulación
      · V_rec           (np array): matriz que contiene los potenciales de membrana neuronales para cada simulation step y para cada
                                    neurona. La matriz está indexada de forma que las filas indican simulation steps y las columnas el
                                    identificador numérico de cada neurona.


    NOTAS DE IMPLEMENTACIÓN
    ------------------------

    La implementación llevada a cabo permite un gran dinamismo y una destacada representatividad biofísica, a costa, claro está, de una
    pobre eficiencia computacional y un estilo de programación de alto nivel linguístico (lo que se conoce como estilo 'pythonico').

    """


    base_path = "/Users/oscar/Library/Mobile Documents/com~apple~CloudDocs/OSCAR/UNIVERSIDAD/FISICA/4º FISICA/TFG/TFG/SIMULATIONS/HEATMAP 8"

    def __init__(self, connective_structure, unineuronal_dynamics):

        """
        Estructura de inicialización de la metaclase Neocortex.

        ARGUMENTOS
        ----------

        La inicialización de la metaclase Neocortex requiere de dos argumentos
          · connective_structure (DiGraph): objeto DiGraph de la librería NetworkX o derivados que contiene la estructura conectiva del
                                            neocortex modelizado en forma de grafo dirigido.
          · unineuronal_dynamics (dic)    : diccionario en el que las claves son los instanciadores de los tipos neuronales que se
                                            consideren (cuidado, no son str) y los valores sus porcentajes en el neocortex modelizado.

        ATRIBUTOS DE INSTANCIA
        -----------------------

        Toda la información relevante a cerca de los atributos de instancia queda recogida en las anotaciones generales de la clase
        Neocortex.

        """

        self.digraph         = connective_structure
        self.digraph_info    = (self.digraph).graph["info"]

        self.model_info      = list(unineuronal_dynamics.keys())[0].model_info

        self.neurons_dic     = unineuronal_dynamics

        self.neocortex       = self.integrate()

        self.dt              = 0.0
        self.simulation_time = 0.0
        self.simulation_long = 0.0

        self.V_rec           = np.array((0,0))


    def integrate (self):

        """
        Integrador de la dinámica unineuronal en la estructura conectiva.

        La integración se lleva a cabo asociando una instanciación neuronal al diccionario de atributos de cada nodo del grafo.

        """

        DG = deepcopy(self.digraph)

        neuron_types     = (self.neurons_dic).keys()
        nodes_for_sample = set(DG.nodes)

        for neuron_type in neuron_types:

            sample_length = len(self.neurons_dic[neuron_type])
            nodes_sample  = rd.sample(nodes_for_sample, sample_length)
            i = 0
            for node in nodes_sample:
                DG.nodes[node]["neuron"] = self.neurons_dic[neuron_type][i]
                i += 1

            nodes_for_sample = nodes_for_sample - set(nodes_sample)

        for node in DG.nodes:
            neuron              = DG.nodes[node]["neuron"]
            neuron.neuron_num   = node
            predeccesor_nodes   = DG.predecessors(node)
            predecessor_neurons = list(map(lambda x: DG.nodes[x]["neuron"], predeccesor_nodes))

            neuron.predecessors = predecessor_neurons

        return(DG)


    def synchrony_coefficient (self, V_matrix = None):

        """
        Calculador del coeficiente de sincronización del neocortex construido

        La sincronización de una estructura neuronal es el grado en el que las neuronas que la componen emiten potenciales de acción
        de manera simultanea en el tiempo, es decir, en una red neuronal perfectamente síncrona todas las neuronas emiten en el mismo
        instante temporal, algo que se traduce en un estado de resonancia inducido por el acoplamiento y la superposición coherente de
        las emisiones.

        Así, un estado de hipersincronía queda caracterizado por grandes fluctuaciones en el potencial de membrana total del sistema.
        Teniendo esto en cuenta se empleará como coeficiente de sincronización la varianza del voltaje promedio del sistema normalizada
        a la unidad. Es decir, se empleará el cociente de la varianza de la media poblacional del voltaje total y la media poblacional
        de las varianzas de los voltajes neuronales individuales.

        ARGUMENTOS
        ----------

        El método para el cálculo del coeficiente de sincronización requiere de un parámetro opcional:
          · V_rec (np array): matriz que contiene los potenciales de membrana neuronales para cada simulation step y para cada neurona.
                              La matriz está indexada de forma que las filas indican simulation steps y las columnas el identificador
                              numérico de cada neurona. El valor por defecrto es self.V_rec que queda definido tras inducir una
                              simulación en la red neuronal.

        SALIDA
        ------

        Devuelve el valor del coeficiente de sincronización (en su interpretación como varianza normalizada) asociado a la distribución
        de voltajes V_matrix.

        """

        if V_matrix == None:
            V_matrix = self.V_rec

        number_neurons   = len(self.neocortex)
        simulation_steps = np.shape(V_matrix)[0]

        enum1 = 0.0
        enum2 = 0.0
        deno1 = np.array([0.0]*number_neurons)
        deno2 = np.array([0.0]*number_neurons)

        for i in range(simulation_steps):
            V_membrane = V_matrix[i]

            enum2 +=  sum(V_membrane)/number_neurons
            enum1 += (sum(V_membrane)/number_neurons)**2
            deno2 +=   V_membrane
            deno1 +=  (V_membrane)**2

        enum =     enum1/simulation_steps - (enum2/simulation_steps)**2
        deno = sum(deno1/simulation_steps - (deno2/simulation_steps)**2)/number_neurons

        synchrony_coefficient = enum/deno

        return(synchrony_coefficient)


    def _simulate_LIF (self):

        """
        Simulador de dinámica de red neuronal en el marco LIF. Se simula una actividad neuronal gobernada por el modelo LIF e integrada
        en la estructura conectiva asociada a la instanciación del neocortex.

        ARGUMENTOS
        ----------

        Aunque el método no admite argumentos explícitos, si que cabe destacar que es necesario definir previamente un intervalo
        temporal de simulación (self.simulation_time) y un parámetro de discretización temporal (self.dt), algo que típicamente se
        lleva a cabo en el método simulate (ver más abajo).

        SALIDA
        ------

        Si bien el método no devuelve explícitamente ningún objeto, como resultado de la simulación se tiene la matriz de potenciales
        de membrana self.V_rec que contiene los valores dicho potencial para cada neurona y cada simulation_step.

        """

        NC                   = self.neocortex
        number_neurons       = len(NC)
        simulation_steps     = round(self.simulation_time/self.dt)

        self.V_rec           = np.zeros((simulation_steps, number_neurons))

        for simulation_step in range(simulation_steps):

            V_membrane = np.array([0.0]*len(NC))

            for node in NC.nodes:
                neuron = NC.nodes[node]["neuron"]
                neuron.avance(simulation_step, self.dt)

                V_membrane[node] = neuron.V_membrane

            self.V_rec[simulation_step, :] = V_membrane.T

            print("Simulation step {}/{}".format(simulation_step + 1, simulation_steps))

        for node in NC.nodes:
            neuron    = NC.nodes[node]["neuron"]
            neuron.has_been_simulated = True


    def save_information(self, directory, neuron_plot = 0, remarcable_info = None):

        """
        Método para el volcado de información en memoria siguiendo diferentes formatos.

        ARGUMENTOS
        ----------

        El método requiere de un argumento indispensable y otro opcional
          · directory       (str): path en el que se va a volcar la información
          · neuron_plot     (int): identificador numérico de la neurona cuyo potencial de membrana quiere graficarse.
          · remarcable_info (str): información relevante que se explicitará en el archivo README.txt asociado al volcado de información
                                   (típicamente debe contener las variaciones o atributos de interés de la simulación cuyos datos van a
                                   ser guardados en memoria)

        SALIDA
        ------

        Si bien el método no devuelve explícitamente ningún objeto, como resultado de su ejecución se crea un directorio cuyo nombre es
        la fecha y hora de su creación (con el fin de evitar posibles conflictos de sobreescritura) y que contiene
          (1) GRÁFICOS         : dos gráficos, uno de tipo raster de las emisiones neuronales y otro  del potencial de membrana de la
                                 neurona neuron_plot, ambos en formato PDF
          (3) GRAFOS           : un archivo del Neocortex modelizado en formato GraphML
          (4) DESERIALIZACIONES: un archivo de la instancia Neocortex en formato pickle (deserialización de forma que una hipotética
                                 serialización permita recuperar íntegramente la instancia)
          (5) MATRICES         : un archivo con la matriz de potenciales de membrana en formato csv (indexada de forma que las filas
                                 indican paso discreto de la simulación y las columnas identificador neuronal numérico)

        NOTAS DE IMPLEMENTACIÓN
        ------------------------
        Se recomienda trabajar con DataFrames para el tratamiento de la matriz de potenciales de membrana. Igualmente, es recomendable
        hacer uso del software Gephi si lo que se quiere es representar y hacer un tratamiento detallado del grafo a través del archivo
        GraphML.

        El archivo pickle es, sin duda, el más potente de todos los anteriores ya que permite recuperar íntegramente la insatancia y
        trabajar con ella a posteriori (graficar patrones, realizar nuevos cálculos, repetir simulaciones variando parámetros ...).
        Para serializar el archivo pickle y cargarlo a memoria priemro será necesario importar el módlulo que contiene la clase
        Neocortex, así como importar explícitamente la propia clase Neocortex. Seguidamente se abrirá el archivo pickle en formato
        lectura binaria para seguidamente cargar a memoria el objeto deserializado, el patrón sería algo así

                                        >>> import modulo
                                        >>> from modulo import Neocortex
                                        >>>
                                        >>> file      = open(directory+/Neocortex_Instance.pickle, "rb")
                                        >>> Neocortex = pickle.load(file)

        """

        NC               = self.neocortex
        simulation_steps = round(self.simulation_time/self.dt)
        number_neurons   = len(NC)

        now  = str(datetime.now().replace(microsecond = 0))

     # (1) MEMBRANE POTENTIAL MATRIX
        V_membrane_df      = DataFrame(data = self.V_rec, index = range(simulation_steps), columns = range(number_neurons))
        V_membrane_df.name = "Membrane Potential Matrix (Rows = simulation steps, Columns = neuron identity number)"
        filename           = directory + "/V_rec.csv"

        V_membrane_df.to_csv(path_or_buf = filename, mode = "w+")

     # (2) README FILE
        filename = directory + "/README.txt"
        readme   = open(filename, "w+")

        readme.write("In this directory some files can be found containing different types of information provided by the simulation carried out on {}\n".format(now))

        readme.write("\n\t· Connective structure: {}".format(self.digraph_info))
        readme.write("\n\t· Unineuronal dynamics: {}".format(self.model_info))
        readme.write("\n\t· Execution time      : {}".format(self.simulation_cost))

        if remarcable_info == None:
            remarcable_info = input("Please, introduce any relevant information about the particularities of the actual simulation: ")
        readme.write("\n\nNOTES: "+remarcable_info)

        readme.close()

     # (3) GRAPH FILE
        NC_str = deepcopy(self)
        for node in NC_str.neocortex.nodes:
            NC_str.neocortex.nodes[node]["neuron"] = NC_str.neocortex.nodes[node]["neuron"].neuron_id
        filename_graphML = directory + "/Neocortex_Graph.graphml"
        filename_pickle  = directory + "/Neocortex_Instance.pickle"
        nx.write_graphml(NC_str.neocortex, path = filename_graphML)
        nx.write_gpickle(self  , path = filename_pickle)

    # (4) PLOT FILES
        figure_raster    = self.plot_raster()

        filename_raster = directory + "/raster_plot.pdf"
        figure_raster.savefig(fname = filename_raster)
        plt.close(figure_raster)

        figure_potential = self.plot_membrane_potential(neuron_number = neuron_plot)
        filename_potential = directory + "/V_membrane_plot.pdf"
        figure_potential.savefig(fname = filename_potential)
        plt.close(figure_potential)


    def plot_raster (self, start_time = 500, end_time = 1000):

        """
        Graficador tipo raster de las emisiones neuronales en el neocortex.

        ARGUMENTOS
        ----------

        El método de graficado del raster plot admite dos argumentos opcionales
          · start_time (float): tiempo a partir del cual quiere observarse la distribución de emisiones neuronales
          · end_time   (float): tiempo hasta el que quiere observarse la distribución de emisiones neuronales.

        SALIDA
        ------

        Se devuelte la figura (objeto de matplotlib) que contiene el raster plot.

        """

        sns.set()
        fig = plt.figure()

        NC             = self.neocortex
        number_neurons = len(NC)

        for node in NC.nodes:
            neuron         = NC.nodes[node]["neuron"]
            emmision_times = neuron.emmision_times
            neuron_number  = [node]*len(emmision_times)
            if   neuron.excitability == "EXCITATORY":
                    plt.scatter(emmision_times, neuron_number, marker = ".", color = sns.xkcd_rgb["pale red"])
            elif neuron.excitability == "INHIBITORY":
                    plt.scatter(emmision_times, neuron_number, marker = ".", color = sns.xkcd_rgb["denim blue"])

        plt.ylim((-1,number_neurons))
        plt.xlabel("Time (ms)")
        plt.ylabel("Neuron ID")
        plt.xlim([start_time,end_time])

        return(fig)



    def plot_membrane_potential(self, start_time = 1000, end_time = None, neuron_number = 0, V_matrix = None):

        """
        Graficador de la variación del potencial de membrana neuronal.

        ARGUMENTOS
        ----------

        El método de graficado del raster plot admite cuatro argumentos opcionales
          · start_time    (float)   : tiempo a partir del cual quiere observarse la variación del potencial de membrana neuronal.
          · end_time      (float)   : tiempo hasta el que quiere observarse la variación del potencial de membrana neuronal.
          · neuron_number (int)     : identificador numérico de la neurona cuyo potencial de membrana se quiere graficar.
          · V_matrix      (np.array): matriz de potenciales de membrana indexada de forma que las filas indiquen pasos discretos de
                                      simulación y las columnas identificadores numéricos neuronales.

        SALIDA
        ------

        Si bien el método no devuelve explícitamente ningún objeto, como resultado de su ejecución se crea un archivo PDF en el
        directorio indicado que contiene la variación del potencial de membrana para la neurona indicada y en base a la última
        simulación a la que haya sido sometido el neocortex.

        """

        if V_matrix == None:
            V_matrix = self.V_rec

        if end_time == None:
            end_time = self.simulation_time

        sns.set()
        fig = plt.figure()

        NC        = self.neocortex
        neuron    = NC.nodes[neuron_number]["neuron"]
        neuron_id = neuron.title_str[:4].upper() + neuron.title_str[4:].title()

        if   neuron.excitability == "EXCITATORY":
            plt.plot(np.arange(0, self.simulation_time, self.dt), V_matrix[:,neuron_number], color = sns.xkcd_rgb["pale red"], label = "Neuron {} ({})".format(neuron_number, neuron_id))
        elif neuron.excitability == "INHIBITORY":
            plt.plot(np.arange(0, self.simulation_time, self.dt), V_matrix[:,neuron_number], color = sns.xkcd_rgb["denim blue"], label = "Neuron {} ({})".format(neuron_number, neuron_id))

        plt.xlabel("Time (ms)")
        plt.ylabel("Voltage (mV)")
        plt.hlines(y=-52.0, xmin = start_time, xmax = end_time, linestyles = "dashed")
        plt.xlim(start_time, end_time)

        return(fig)


    def simulate (self, model = "LIF", simulation_time = 1500.0, time_step = 0.05, save_info = True, remarcable_info = None):

        """
        Simulador de dinámica de red neuronal en cualquier marco de dinámica unineuronal modelizado. Se simula una actividad neuronal
        gobernada por el modelo escogido e integrada en la estructura conectiva asociada a la instanciación del neocortex.

        ARGUMENTOS
        ----------

        El método de simulación general tiene 5 argumentos opcionales
          · model           (str)  : modelo de dinámica unineuronal escogido para la simulación.
          · simulation_time (float): tiempo propio del neocortex durante el que se desea llevar a cabo la simulación.
          · time_step       (float): parámetro de discretización temporal empleado en la simulación.
          · save_info       (bool) : booleano que indica si se desea llamar al método save_information y, por tanto, realizar un
                                     volcado estandar de los datos de la simulación.
          · remarcable_info (str)  : información relevante que se explicitará en el archivo README.txt asociado al volcado de información
                                     (típicamente debe contener las variaciones o atributos de interés de la simulación cuyos datos van a
                                     ser guardados en memoria)

        """

        self.simulation_time = simulation_time
        self.dt              = time_step
        if model.upper()[0] == "L":
            start_time = datetime.now()
            self._simulate_LIF()
            end_time   = datetime.now()

        self.simulation_cost = end_time - start_time

        if save_info:
            now  = str(datetime.now().replace(microsecond = 0))
            path = self.base_path + "/" + now

            os.mkdir(path)
            self.save_information(directory = path, remarcable_info = remarcable_info)


    def __repr__ (self):

        """
        Representación formal del neocortex.

        """
        return("Neocortex")


    def __str__ (self):

        """
        Representación en forma de cadena de caracteres del neocortex, en la que se detalla toda su información relevante.

        """

        string_title= "-------------------------------------\nHUMAN NEOCORTICAL MODEL\n------------------------------------- \
        \nThis is a human neocortex, product of the proyection of an unineuronal dynamic to a certain connective strucute:"

        string_unineuronal_dinamics = "\
        \n\n\t(1) UNINEURONAL DINAMICS \
        \n\t    --------------------\
        \n\t    The human neocortex is known to contain both \
        \n\t\t   - NEOCORTICAL PIRAMIDAL NEURONS: glutamatergic (excitatory) proyection neurons (80% of neocortical neurons for a physiological adult human brain) \
        \n\t\t   - NEOCORTICAL INTERNEURONS     : GABAergic inhibitory intrinsic neurons        (20% of neocortical neurons for a physiological adult human brain) \
        \n\t    which have been modeled in the {0} framework. In this actual neocortex the neuron distribution is as follows:".format(self.model_info)

        string_actual_unineuronal_dinamics = ""
        for neuron in self.neurons_dic.keys():
            string_actual_unineuronal_dinamics += "\
        \n\t\t   - {0:6.2f}% {1}".format(self.neurons_dic[neuron], neuron.title_str[3:])

        string_connective_structure = "\
        \n\n\t(2) CONNECTIVE STRUCTURE \
        \n\t    --------------------\
        \n\t    The neocortical conective structure has been modeled using a {0}\n".format(self.digraph_info)

        return(string_title+string_unineuronal_dinamics+string_actual_unineuronal_dinamics+string_connective_structure)


#--------------------------------------------------------------------------------------------------------------------------------------
# PRUEBAS
#--------------------------------------------------------------------------------------------------------------------------------------

if __name__ == "__main__":

    number_neurons = 100
    neighbours     = 10
    p              = 0.2

    percentage_P   = 80.0
    percentage_I   = 20.0
    number_neurons_P = round(number_neurons*(percentage_P/100.0))
    number_neurons_I = round(number_neurons*(percentage_I/100.0))

    columns = ["SC"]
    index   = [round(x, 2) for x in np.arange(0.0, 8.8, 0.1)]
    data1   = DataFrame(columns = columns, index = index)

    print(data1)
    filename = "/Users/oscar/Library/Mobile Documents/com~apple~CloudDocs/OSCAR/UNIVERSIDAD/FISICA/4º FISICA/TFG/TFG/SIMULATIONS/CANNABIS 2/cannabis_synchro.csv"

    i = 1
    for g in data1.index:

        DiWS = DiWattsStrogatz(number_neurons, neighbours, p)

        parameter_dic = {
        "P_NC" : {
          "time_refractory"           :   2.0,
          "tau_membrane"              :  20.0,
          "V_threshold"               : -52.0,
          "g_membrane"                :  26.3,
          "V_rest"                    : -70.0,
          "V_membrane_initial_range"  : (-80.0,-55.0),
          "I_ext_range"               : (490.0,510.0)},
        "I_NC" : {
          "time_refractory"           :   1.0,
          "tau_membrane"              :  10.0,
          "V_threshold"               : -52.0,
          "g_membrane"                :  26.3,
          "V_rest"                    : -70.0,
          "V_membrane_initial_range"  : (-80.0,-55.0),
          "I_ext_range"               : (490.0,510.0)}}

        receptor_dic = {
        "P_NC" : {
          "IONTROPIC"    : {"EXCITATORY" : {"AMPA" : {"V_syn"             :   0.0,
                                                      "tau_r"             :   0.5,
                                                      "tau_d"             :   2.0,
                                                      "conduction_delay"  :   0.0,
                                                      "coupling_strength" :   2.5}},
                            "INHIBITORY" : {"GABA" : {"V_syn"             : -80.0,
                                                      "tau_r"             :   0.5,
                                                      "tau_d"             :   5.0,
                                                      "conduction_delay"  :   0.0,
                                                      "coupling_strength" :  (11.1/8.8)*g}}},
          "METABOTROPIC" : {}},
        "I_NC" : {
          "IONTROPIC"    : {"EXCITATORY" : {"AMPA" : {"V_syn"             :   0.0,
                                                      "tau_r"             :   0.5,
                                                      "tau_d"             :   2.0,
                                                      "conduction_delay"  :   0.0,
                                                      "coupling_strength" :   2.0}},
                            "INHIBITORY" : {"GABA" : {"V_syn"             : -80.0,
                                                      "tau_r"             :   0.5,
                                                      "tau_d"             :   5.0,
                                                      "conduction_delay"  :   0.0,
                                                      "coupling_strength" :   g}}},
          "METABOTROPIC" : {}}}

        unineuronal_dynamics = {Piramidal_NC : number_neurons_P, Interneuron_NC : number_neurons_I}

        neuron_types = unineuronal_dynamics.keys()
        for neuron_type in neuron_types:
            number_neurons_create             = unineuronal_dynamics[neuron_type]
            unineuronal_dynamics[neuron_type] = [neuron_type(general_parameter_dic = parameter_dic, general_receptor_dic = receptor_dic) for _ in range(number_neurons_create)]

        N = Neocortex(connective_structure = DiWS, unineuronal_dynamics = unineuronal_dynamics)

        remarcable = "Esta simulación forma parte de una serie de simulaciones cuyo objetivo es realizar un MODELO DE NEOCORTEX BAJO EFECTO DE THC EN CONCRETO g={}".format(g)
        N.simulate(model = "LIF", simulation_time = 1500.0, time_step = 0.05, save_info = True, remarcable_info = remarcable)

        data1.loc[g, "SC"] = N.synchrony_coefficient()
        data1.to_csv(path_or_buf = filename, mode = "w+")


        print("\n\n\n\n\n\n\n SIMULATION {}/{} COMPLETED \n\n\n\n\n\n\n".format(i, len(data1.index)*len(data1.columns)))
        i +=1
