"""
---------------------------------------------------------------------------------------------------------------------------------------
MÓDULO NEURON
---------------------------------------------------------------------------------------------------------------------------------------

El módulo neuron dispone de la metaestructura OOP necesaria para introducir modelos de dinámica unineuronal y consta de 4 niveles de
profundidad jerárquica.

Se ha implementado completamente el modelo LIF (Leaky Integrate and Fire), concretado en dos tipos neuronales específicos: neuronas
piramidales glutamatérgicas (excitatorias) e interneuronas GABAergicas (inhibitorias), ambas propias del tejido neocortical humano.

La estructura ha sido diseñada para facilitar incorporaciones y adiciones futuras de nuevos modelos y elementos.

"""

import numpy as np
import scipy as sp

from collections     import defaultdict
from copy            import deepcopy

#--------------------------------------------------------------------------------------------------------------------------------------
# GENERAL LAMBDAS
#--------------------------------------------------------------------------------------------------------------------------------------

recursive_dictionary = lambda : defaultdict(recursive_dictionary)

#--------------------------------------------------------------------------------------------------------------------------------------
# GENERAL FUNCTIONS
#--------------------------------------------------------------------------------------------------------------------------------------

def default_to_regular(d):

    """
    Conversor de arbol de default_dict anidados a usual dict, si d no es un defaultdict devuelve d inmutado.

    """
    if isinstance(d, defaultdict):
        d = {k: default_to_regular(v) for k, v in d.items()}
    return d

#--------------------------------------------------------------------------------------------------------------------------------------
# METACLASE CELULAR DE PRIMER ORDEN (PRIMERA GENERACIÓN PARENTAL)
#--------------------------------------------------------------------------------------------------------------------------------------

class Cell():

    """
    Metaclase de primera generación (orden 1) para la la modelización celular cuya única pretensión es vertebrar el esquema de
    programación orientada a objetos escogido.

    """

    pass

#--------------------------------------------------------------------------------------------------------------------------------------
# METACLASES CELULARES DE SEGUNDO ORDEN (SEGUNDA GENERACIÓN PARENTAL)
#--------------------------------------------------------------------------------------------------------------------------------------

class Neuron(Cell):

    """
    Metaclase de segunda generación (orden 2) para la modelización de células neuronales cuya única pretensión es aportar un subnivel
    de vertebrado el esquema de programación orientada a objetos escogido.

    """

    pass

class Neuroglia(Cell):

    """
    Metaclase de segunda generación (orden 2) para la modelización de células gliales cuya única pretensión es aportar un subnivel
    de vertebrado el esquema de programación orientada a objetos escogido.

    """

    pass

#--------------------------------------------------------------------------------------------------------------------------------------
# METACLASES CELULARES DE TERCER ORDEN (TERCERA GENERACIÓN PARENTAL)
#--------------------------------------------------------------------------------------------------------------------------------------

class LIFN (Neuron):

    """
    Leaky Integrate and Fire Neuron (LIFN). Metaclase de tercera generación (orden 3) para la modelización neuronal que contiene toda
    la metaestructura relativa a las unidades neuronales representadas por un modelo LIF (Leaky Integrate and Fire).

    El LIF es un modelo neuronal con representatividad biofísica que destaca por su simplicidad. En concreto forma parte del grupo
    PND (passive neuronal dynamics) que no aspiran a dar parte de la forma ni la evolución de los potenciales de acción interneuronales.

    La aproximación PND se justifica teniendo en cuenta que todos los potenciales de acción interneuronales son aproximadamente iguales
    en términos morfológicos por lo que la forma del pulso no debería contener la información transmitida.

    Así, los modelos PND introducen un marco binario en el que se consideran los potenciales de acción como eventos cuya (no) ocurrencia
    es la unidad de información (algo que recuerda, no sin razón, a los propios cimientos de la electrónica digital).

    ATRIBUTOS DE CLASE
    ------------------

    Los atributos de clase se erigen de forma que algunos permanecerán inmutados en las clases descendientes y otros actuarán como
    estructuras paradigmáticas (contenedores formales) que serán dotadas de contenido dentro de las clases descendientes.
      · default_parameter_dic (dic): diccionario que contiene los parámetros neuronales por defecto.
      · default_receptors_dic (dic): diccionario que contiene los parámetros por defecto de los receptores.
      · model_info            (str): cadena con la información a cerca del modelo biofísico al que pertenece la neurona
      · title_str             (str): cadena con el título utilizado en la representación str de la clase
      · neuron_type_str       (str): cadena con el nombre del tipo neuronal que será utilizada en la representación str de la clase
      · excitatority          (str): cadena con el carácter de la neurona (excitatorio/inhibitorio)
      · neuron_id             (str): cadena con el identificador del tipo neuronal


    ATRIBUTOS DE INSTANCIA
    ----------------------

    Por un lado se definen una serie de parámetros inherentes a cualquier neurona (ya sea aislada o integrada en ua red neuronal)
      · general_parameters_dic     (dic)  : diccionario que contiene los parámetros generales de todos los tipos neuronales considerados
      · general_receptors_dic      (dic)  : diccionario que contiene los parámetros generales de todos los receptores considerados
      · self_parameters_dic        (dic)  : diccionario que contiene los parámetros particulares de la neurona modelizada.
      · self_receptors_dic         (dic)  : diccionario que contiene los parámetros de los receptores de la neurona modelizada
      · postsynaptic_receptors_dic (dic)  : diccionario que contiene los parámetros generales de los receptores diana de la neurona
      · general_neuron_ides        (list) : lista con todos los identificadores de tipo neuronal considerados
      · tau_membrane               (float): constante de tiempo de la membrana
      · time_refractory            (float): tiempo en periodo refractario
      · V_rest                     (float): potencial de reposo
      · V_threshold                (float): potencial umbral
      · g_membrane                 (float): conductancia de la membrana
      · R_membrane                 (float): resistencia de la membrana
      · I_ext                      (float): corriente externa inyectada
      · V_membrane                 (float): potencial de membrana

    Igualmente, se define un segundo conjunto de parámetros que únicamente serán relevantes cuando la neurona es integrada en una
    estructura conectiva
      · dt                         (float): parámetro de discretización temporal
      · I_syn                      (float): valor de la corriente sináptica
      · predecessors               (list) : lista con las neuronas presináptica de la nerona dada
      · emmision_times             (list) : lista con los tiempos en los que la neurona ha emitido
      · has_been_simulated         (bool) : indica si la neurona ha completado una simulación o no
      · emmited                    (bool) : evaludado en un simulation_step refleja si la neurona ha emitido o no
      · emmited_conductance_pulse  (dic)  : diccionario con la información de los pulsos de conductancia que emite la neurona.
      · last_spikes                (dic)  : diccionario con las emisiones consideradas para calcular el pulso de conductancia saliente
      · simulation_step            (int)  : indica el step de la simulación en el que se encuentra la neurona.
      · neuron_num                 (int)  : entero con el nodo que posiblemente ocupe la neurona en una estructura conectiva

    NOTAS DE IMPLEMENTACIÓN
    ------------------------

    La implementación llevada a cabo permite un gran dinamismo y una destacada representatividad biofísica, a costa, claro está, de una
    pobre eficiencia computacional y un estilo de programación de alto nivel linguístico (lo que se conoce como estilo 'pythonico').

    Cabe destacar también que se ha elegido dotar de grán generalidad a las metaclases generadoras de los marcos de dinámica unineuronal
    (LIFN para Leaky Integrate and Fire, HHN para Hodking-Huxley, ...) de forma que la materialización de los diferentes tipos neuronales
    resulte directa y clara y contenga únicamente atributos diferenciales.

    """

    default_parameter_dic = {
    "P_NC" : {
      "time_refractory"           :   2.0,
      "tau_membrane"              :  20.0,
      "V_threshold"               : -52.0,
      "g_membrane"                :  26.3,
      "V_rest"                    : -70.0,
      "V_membrane_initial_range"  : (-80.0,-55.0),
      "I_ext_range"               : (490.0,510.0)},
    "I_NC" : {
      "time_refractory"           :   1.0,
      "tau_membrane"              :  10.0,
      "V_threshold"               : -52.0,
      "g_membrane"                :  26.3,
      "V_rest"                    : -70.0,
      "V_membrane_initial_range"  : (-80.0,-55.0),
      "I_ext_range"               : (490.0,510.0)}}

    default_receptors_dic = {
    "P_NC" : {
      "IONTROPIC"    : {"EXCITATORY" : {"AMPA" : {"V_syn"             :   0.0,
                                                  "tau_r"             :   0.5,
                                                  "tau_d"             :   2.0,
                                                  "conduction_delay"  :   0.0,
                                                  "coupling_strength" :   2.5}},
                        "INHIBITORY" : {"GABA" : {"V_syn"             : -80.0,
                                                  "tau_r"             :   0.5,
                                                  "tau_d"             :   5.0,
                                                  "conduction_delay"  :   0.0,
                                                  "coupling_strength" :  11.1}}},
      "METABOTROPIC" : {}},
    "I_NC" : {
      "IONTROPIC"    : {"EXCITATORY" : {"AMPA" : {"V_syn"             :   0.0,
                                                  "tau_r"             :   0.5,
                                                  "tau_d"             :   2.0,
                                                  "conduction_delay"  :   0.0,
                                                  "coupling_strength" :   2.0}},
                        "INHIBITORY" : {"GABA" : {"V_syn"             : -80.0,
                                                  "tau_r"             :   0.5,
                                                  "tau_d"             :   5.0,
                                                  "conduction_delay"  :   0.0,
                                                  "coupling_strength" :   8.8}}},
      "METABOTROPIC" : {}}}

    model_info      = "Leaky Integrate and Fire (LIF)"
    title_str       = "LIF GENERAL NEURON"
    neuron_type_str = "general neuron"
    excitability    = "GENERAL"
    neuron_id       = "GENERAL"
    neuron_num      = "NEURON NUMBER NOT DEFINED!"


    def __init__ (self, general_parameter_dic, general_receptor_dic):

        """
        Estructura de inicialización de la metaclase LIFN. Se trata del paradigma de inicialización que será heredado por las clases
        hijas para la inicialización de los diferentes tipos neuronales dentro del marco LIF.

        ARGUMENTOS
        ----------

        La inicialización de la metaclase LIFN requiere de dos argumentos
          - general_parameter_dic (dic): diccionario que contiene los parámetros generales de todos los tipos neuronales considerados
          - general_receptor_dic  (dic): diccionario que contiene los parámetros generales de todos los receptores considerados

        ATRIBUTOS DE INSTANCIA
        -----------------------

        Toda la información relevante a cerca de los atributos de instancia queda recogida en las anotaciones generales de la clase
        LIFN, no obstante, cabe remarcar que se diferencian dos tipos de atributos de instancia:
          (1) Los inherentes a cualquier neurona (aislada o integrada en una estructura conectiva)
          (2) Los que únicamente trascienden y quedan definidos cuando la dinámica unineuronal es proyectada a una estructura conectiva

        NOTAS DE IMPLEMENTACIÓN
        -----------------------

        Referencia (*): la transformación del objeto <dic_keys> a <list> responde a una peculiar y poca conocida carencia de la
        librería pickle que será utilizada más adelante para guardar en memoria instancias de diferentes clases. Esta carencia está
        ligada a la transición entre las versiones 2 y 3 de Python y hace imposible deserializar, por ejemplo, objetos tipo <dic_keys>
        o <dic_values>. El problema es tán engorroso de solventar a nivel de la librería pickle que la directriz general es no
        emplear este tipo de estructuras si se pretende trabajar con serializaciones y deserializaciones en el marco de la programación
        orientada a objetos.

        """


      # (1) ATRIBUTOS GENERALES

        self.general_receptors_dic      = general_receptor_dic
        self.general_parameters_dic     = general_parameter_dic

        self.self_receptors_dic         = self.general_receptors_dic[self.neuron_id]
        self.self_parameters_dic        = self.general_parameters_dic[self.neuron_id]

        self.general_neuron_ides        = list(self.general_parameters_dic.keys()) # (*) necesario para que pickle no chille por incompatibilidades python 2 - 3 (picklear objetos tipo dict_keys no posible)

        self.tau_membrane               = self.self_parameters_dic["tau_membrane"]
        self.time_refractory            = self.self_parameters_dic["time_refractory"]

        self.V_rest                     = self.self_parameters_dic["V_rest"]
        self.V_threshold                = self.self_parameters_dic["V_threshold"]

        self.g_membrane                 = self.self_parameters_dic["g_membrane"]
        self.R_membrane                 = 1.0/self.g_membrane

        self._I_ext_range               = self.self_parameters_dic["I_ext_range"]
        self._V_membrane_initial_range  = self.self_parameters_dic["V_membrane_initial_range"]

        self.I_ext                      = np.random.uniform(*self._I_ext_range)
        self.V_membrane_initial         = np.random.uniform(*self._V_membrane_initial_range)
        self.V_membrane                 = self.V_membrane_initial

        self.postsynaptic_receptors_dic = deepcopy(self.general_receptors_dic)
        if   self.excitability == "INHIBITORY": excitability_to_delete = "EXCITATORY"
        elif self.excitability == "EXCITATORY": excitability_to_delete = "INHIBITORY"
        for neuron_id in self.general_neuron_ides:
            try:
                del self.postsynaptic_receptors_dic[neuron_id]["IONTROPIC"][excitability_to_delete]
            except KeyError:
                pass
            receptors = self.postsynaptic_receptors_dic[neuron_id]["IONTROPIC"][self.excitability]
            self.postsynaptic_receptors_dic[neuron_id]["IONTROPIC"] = receptors

      # (2) ATRIBUTOS ESPECÍFICOS
        self.predecessors               = []

        self.dt                         = 1.0
        self.simulation_step            = 0

        self.I_syn                      = 0.0

        self.emmited                    = False
        self.emmision_times             = []
        self.emmited_conductance_pulse  = {}

        self.has_been_simulated         = False

        self.last_spikes                = recursive_dictionary()
        for target_neuron_id in self.general_neuron_ides:
            for target_receptor_id in self.postsynaptic_receptors_dic[target_neuron_id]["IONTROPIC"].keys():
                self.last_spikes[target_neuron_id][target_receptor_id] = []
        self.last_spikes                = default_to_regular(self.last_spikes)


    @property
    def I_ext_range (self):

        """
        Redefinición mediante un decorator del rango de intensidades externas a partir del que aleatoriamente se le asigna una
        intensidad externa a la neurona cuando esta es inicializada.

        """
        return self._I_ext_range

    @I_ext_range.setter
    def I_ext_range (self, I_range):

        """
        Asignador de valor al rango de intensidades externas.

        La concepción del rango de intensidades externas como una propiedad y no como atributo permite un mayor versatilidad ya que de
        este modo la modificación del rango de intensidades una vez instanciada la clase automáticamente induce una modificación en el
        valor de la intensidad externa de la neurona en el rango indicado.

        ARGUMENTOS
        -----------

        El método de asignación de la propiedad I_ext_range requiere de un único argumento:
          · I_range (tuple): tupla con dos elementos indicando el rango de valores a partir del que se le asignará de forma aleatoria
                             una intensidad externa a la neurona.

        EJEMPLOS
        --------

        Una asignación típica puede ser la siguiente

          >>> L = LIFN(parameter_dic, receptor_dic)
          >>> L.I_ext_range = (450.0, 470.0)
          >>> L.I_ext
          467.3

        """

        self.self_parameters_dic["I_ext_range"] = I_range
        self._I_ext_range                       = self.self_parameters_dic["I_ext_range"]
        self.I_ext                              = np.random.uniform(*self._I_ext_range)


    @property
    def V_membrane_initial_range (self):

        """
        Redefinición mediante un decorator del rango de potenciales de membrana iniciales a partir del que aleatoriamente se le asigna
        un potencial de membrana a la neurona cuando esta es inicializada.

        """
        return self._V_membrane_initial_range

    @V_membrane_initial_range.setter
    def V_membrane_initial_range (self,V_membrane_initial_range):

        """
        Asignador de valor al rango de intensidades externas.

        La concepción del rango de potenciales de membrana iniciales como  propiedad y no como atributo permite un mayor dinamismo ya
        que de este modo la modificación del rango de potenciales una vez instanciada la clase automáticamente induce una modificación
        en el valor del potencial de membrana inicial de la neurona en el rango indicado.

        ARGUMENTOS
        -----------

        El método de asignación de la propiedad V_membrane_initial_range requiere de un argumento:
          · V_m_initial_range (tuple): tupla con 2 elementos indicando el rango de valores a partir del que se le asignará de forma
                                       aleatoria un potencial de membrana inicial a la neurona.

        EJEMPLOS
        --------

        Una asignación típica puede ser la siguiente (Piramidal_NC es una clase hija de LIFN, ver el apartado de clases neuronales
        derivadas de la clase LIFN)

          >>> P = Piramidal_NC()
          >>> P.V_membrane_initial_range = (-58.0, -80.0)
          >>> P.V_membrane_initial
          62.4

        """

        self.self_parameters_dic["V_membrane_initial_range"] = V_membrane_initial_range
        self._V_membrane_initial_range                       = self.self_parameters_dic["V_membrane_initial_range"]
        self.V_membrane_initial                              = np.random.uniform(*self._V_membrane_initial_range)
        self.V_membrane                                      = self.V_membrane_initial


    def add_receptor(self, receptor_type, receptor_name, info_dic, excitability = None):

        """
        Adicionador de receptores sinápticos a la neurona.

        Si el receptor es iontrópico tendrá una excitabilidad bien definida por lo que esta deberá ser especificada. Además, en el caso
        iontrópico el info_dic deberá contener valores para los parámetros que en el modelo LIF caracterizan un receptor, es decir,
        deberá ser de la forma

                                                     {"V_syn"             : ___,
                                                      "tau_r"             : ___,
                                                      "tau_d"             : ___,                                                  (*)
                                                      "conduction_delay"  : ___,
                                                      "coupling_strength" : ___}

        ARGUMENTOS
        ----------

        El método de adición de receptores requiere de 3 argumentos para receptores metabotrópicos y de 4 para receptores iontrópicos
          · receptor_type (str): cadena de caracteres que debe ser "IONTROPIC" si el receptor es iontrópico y "METABOTROPIC" si el
                                 receptor es metabotrópico.
          · receptor_name (str): cadena de caracteres que indica el nombre del receptor (tener en cuenta que ese nombre será utilizado
                                 para referenciar el receptor y sus propiedades a lo largo del código)
          · info_dic      (dic): diccionario que contiene las propiedades y parámetros que definen el receptor en el marco del modelo
                                 Leaky Integrate and Fire (LIF).
          · excitability  (str): cadena que únicamente debe ser expecificada si el receptor es de tipo iontrópico y que debe ser
                                 "EXCITATORY" si el receptor es de carácter excitatorio (receptor de neurotransmisores excitatorios) e
                                 "INHIBITORY" si el receptor es de carácter inhibitorio (receptor de neurotransmisores inhibitorios)

        EJEMPLOS
        --------

        Una creación típica de un receptor puede ser la siguiente (Piramidal_NC es una clase hija de LIFN, ver el apartado de clases
        neuronales derivadas de la clase LIFN)

          >>> P = Piramidal_NC()
          >>> info = {Vsyn = -20.0,coupling_strength = 0.3, conduction_delay = 0.0, tau_r = 0.3, tau_d = 2}
          >>> P.add_receptor(receptor_type = "IONTROPIC", excitability = "EXCITATORY",
                             receptor_name = "NDMA", info_dic = info)

        NOTAS DE IMPLEMENTACIÓN
        -----------------------

        Según está implementado el método también permite editar receptores mediante sobreescritura, es decir, que el mismo método
        puede referirse a un receptor existente y modificar uno o varios de sus parámetros modelísticos.

        """
        if receptor_type == "IONTROPIC":
            self.self_receptors_dic[receptor_type][excitability][receptor_name] = info_dic
        else:
            self.self_receptors_dic[receptor_type][receptor_name]               = info_dic


    def delete_receptor(self, receptor_type, receptor_name, excitability = None):

        """
        Eliminador de receptores sinápticos de la neurona.

        ARGUMENTOS
        ----------

        El método de eliminación de receptores requiere de 2 argumentos para receptores metabotrópicos y de 3 para iontrópicos
          · receptor_type (str): cadena de caracteres que debe ser "IONTROPIC" si el receptor es iontrópico y "METABOTROPIC" si el
                                 receptor es metabotrópico.
          · receptor_name (str): cadena de caracteres que indica el nombre del receptor (tener en cuenta que ese nombre será utilizado
                                 para referenciar el receptor y sus propiedades a lo largo del código)
          · excitability  (str): cadena que únicamente debe ser expecificada si el receptor es de tipo iontrópico y que debe ser
                                 "EXCITATORY" si el receptor es de carácter excitatorio (receptor de neurotransmisores excitatorios) e
                                 "INHIBITORY" si el receptor es de carácter inhibitorio (receptor de neurotransmisores inhibitorios)

        EJEMPLOS
        --------

        Una eliminación típica de un receptor puede ser la siguiente (Piramidal_NC es una clase hija de LIFN, ver el apartado de clases
        neuronales derivadas de la clase LIFN)

          >>> P = Piramidal_NC()
          >>> P.delete_receptor(receptor_type = "IONTROPIC", excitability = "EXCITATORY",
                                receptor_name = "GABA")

        """

        if receptor_type == "IONTROPIC":
            del self.self_receptors_dic[receptor_type][excitability][receptor_name]
        else:
            del self.self_receptors_dic[receptor_type][receptor_name]


    def conductance_pulse_function(self, t, g, tau_d, tau_r):

        """
        Función pulso de conductancia modelizado como diferencia de exponenciales.

        ARGUMENTOS
        ----------

        El método pulso de conductancia requiere de 3 parámetros (específicos del tipo neuronal que y del receptor sináptico activado)
        que definen la forma del pulso y una variable (temporal)
          · g     (float): intensidad de acoplamiento sináptico en nS (nano-siemens)
          · tau_d (float): tiempo de decaimiento del pulso ms (mili-segundos)
          · tau_r (float): tiempo de subida del pulso en ms (mili-segundos)

          · t     (float): tiempo en el que quiere evaluarse el pulso, también puede ser una lista de tiempos (Python tiene un buen
                           tratamiento álgebra vectorial)

        SALIDA
        ------

        La función devuelve la evaluación del punto de conductancia en el instante t.

        """

        return(g*(np.exp(-t/tau_d)-np.exp(-t/tau_r)))


    def emmited_conductance_pulse_calc(self):

        """
        Conjunto de valores representativos del conductancia inducido sobre los diferentes recptores de las diferentes neuronas.

        SALIDA
        ------

        Devuelve un diccionario anidado que para cada tipo de receptor (excitable por la neurona presináptica) en cada tipo de neurona
        postsináptica contiene un array con los valores del pulso de conductancia (en nano-siemens nS) inducido sobre la neurona
        postsináptica indexados por el número de simulation steps transcurridos desde la emisión del pulso.

        La estructura anidada se justifica teniendo en cuenta que el pulso de conductancia inducido depende no solo de la neurona
        "diana" sino también del receptor activado.

        NOTAS DE IMPLEMENTACIÓN
        -----------------------

        El número máximo de simulation steps que se consideran son 6 veces el tiempo de decaimiento asociado al receptor neuronal (a
        partir de ese momento las contribuciones se consideran despreciables).

        """

        conductance_pulse      = recursive_dictionary()

        for target_neuron_id in self.general_neuron_ides:
            for target_receptor_id in self.postsynaptic_receptors_dic[target_neuron_id]["IONTROPIC"].keys():
                dic   = self.general_receptors_dic[target_neuron_id]["IONTROPIC"][self.excitability][target_receptor_id]

                tau_r = dic["tau_r"]
                tau_d = dic["tau_d"]
                g     = dic["coupling_strength"]

                net_times = np.arange(0, 6*tau_d, self.dt)
                values    = self.conductance_pulse_function(net_times, g, tau_d, tau_r)
                conductance_pulse[target_neuron_id][target_receptor_id] = values

        conductance_pulse      = default_to_regular(conductance_pulse)

        return(conductance_pulse)


    def g_out(self, simulation_step):

        """
        Valor neto del pulso de conductancia inducido sobre los diferentes receptores de las diferentes neuronas en el simulation step
        indicado.

        ARGUMENTOS
        ----------

        El método pulso de conductancia neto requiere de un único argumento:
          · simulation_step (int) : paso discreto de simulación en el que quiere computarse el valor neto del pulso inducido.

        SALIDA
        ------

        Devuelve un diccionario anidado que para cada tipo de receptor (excitable por la neurona presináptica) en cada tipo de neurona
        postsináptica contiene el valor neto del pulso de conductancia (en nano-siemens nS) inducido sobre la neurona postsináptica en
        el simulation step indicado.

        La estructura anidada se justifica teniendo en cuenta que el pulso de conductancia inducido depende no solo de la neurona
        "diana" sino también del receptor activado.

        """

        g_out = recursive_dictionary()

        conductance_pulse = self.emmited_conductance_pulse

        for target_neuron_id in self.general_neuron_ides:
            for target_receptor_id in self.postsynaptic_receptors_dic[target_neuron_id]["IONTROPIC"].keys():

                g_out[target_neuron_id][target_receptor_id] = 0.0

                dic = self.general_receptors_dic[target_neuron_id]["IONTROPIC"][self.excitability][target_receptor_id]

                conduction_delay = dic["conduction_delay"]
                tau_d            = dic["tau_d"]

                steps_conduction_delay = int(round(conduction_delay/self.dt))
                steps_for_dismishing   = int(round(6*tau_d/self.dt))

                for j in range(len(self.last_spikes[target_neuron_id][target_receptor_id])):

                    time_step = simulation_step - self.last_spikes[target_neuron_id][target_receptor_id][j] # (*)

                    if (time_step >= steps_conduction_delay and time_step < steps_for_dismishing + steps_conduction_delay):
                        conductance_pulse_array = conductance_pulse[target_neuron_id][target_receptor_id]

                        g_out[target_neuron_id][target_receptor_id] += conductance_pulse_array[time_step - steps_conduction_delay]

        g_out = default_to_regular(g_out)

        return(g_out)


    def I_syn_calculate(self, simulation_step, V):

        """
        Valor de la corriente sináptica inducida en la neurona en el simulation step indicado cuando su potencial de membrana es V.

        ARGUMENTOS
        ----------

        El método calculador de la corriente sináptica requiere de dos argumentos:
          · V               (float): potencial de membrana neuronal para el que quiere computarse el valor de la corriente sináptica
          · simulation_step (int)  : paso discreto de simulación para el que quiere computarse el valor de la corriente sináptica

        SALIDA
        ------

        Devuelve el valor de la corriente sináptica inducida en la nuerona (en pico-amperios pA) como consecuencia de los pulsos
        emitidos por todas sus neuronas predecesoras en el simulation step dado y cuando el potencial de membrana de la neurona tiene
        el valor proporcionado.

        Igualmente, el método actualiza el valor del atributo I_syn cada vez que es llamado.

        """

        I_syn = 0.0
        for presynaptic_neuron in self.predecessors:
            for postsynaptic_receptor_id in presynaptic_neuron.postsynaptic_receptors_dic[self.neuron_id]["IONTROPIC"].keys():

                g_out = presynaptic_neuron.g_out(simulation_step)[self.neuron_id][postsynaptic_receptor_id]
                V_syn = self.self_receptors_dic["IONTROPIC"][presynaptic_neuron.excitability][postsynaptic_receptor_id]["V_syn"]

                I_syn += g_out * (V_syn - V)

        self.I_syn = I_syn

        return(I_syn)


    def differential_equation_function(self, simulation_step, V):

        """
        Función derivada temporal del potencial de membrana en el simulation step indicado cuando su potencial de membrana es V. La ED
        que gobierna la dinámica unineuronal en el marco LIF

                             C(dVm/dt) = Irest(t) + Isyn(t) + Iext

        Puede reescribirse de forma que resulte en la expresión implementada en el método.

                  dVm/dt = -[V(t) - V_rest] - Rm * g_syn(t) [V(t)-Vsyn] + Rm * I_ext

        ARGUMENTOS
        ----------

        El método calculador de la corriente sináptica requiere de dos argumentos:
          · V               (float): potencial de membrana neuronal para el que quiere computarse el valor de la corriente sináptica
          · simulation_step (int)  : paso discreto de simulación para el que quiere computarse el valor de la corriente sináptica

        SALIDA
        -------

        Devuelve el valor de la derivada temporal del potencial de membrana en el simulation step indicado cuando su potencial de
        membrana es V.

        """

        Isyn = self.I_syn_calculate(simulation_step = simulation_step, V = V)

        return (-(V - self.V_rest) + self.R_membrane * Isyn + self.R_membrane * self.I_ext) / self.tau_membrane


    def avance(self, simulation_step, dt):

        """
        Inducidor de avance en una unidad de discretización temporal sobre la neurona en el marco de una simulación.

        Aspira a modelizar, en el marco LIF, el resultado formal de la actividad citológico-fisiológica de una neurona integrada en una
        estructura plurineuronal.

        ARGUMENTOS
        ----------

        El método de avance requiere de dos argumentos:
          · simulation_step (int)  : paso discreto de la simulación en la que se encuentra el sistema en global (no confundir con
                                     self.simulation_step, ver NOTAS DE IMPLEMENTACIÓN)
          · dt              (float): unidad de discretizción temporal empleada en la simulación

        NOTAS DE IMPLEMENTACIÓN
        -----------------------

        Resulta trascendental no confundir ni identificar ciegamente las cantidades simulation_step (argumento de la función) y
        self.simulation_step (atributo de la instancia).
          ·      simulation_step : es el paso discreto de la simulación en el que, de forma global, se encuentra el sistema simulado.
          · self.simulation_step : es el paso discreto de la simulación en el que, de forma concfreta, se encuentra la neurona.
        Debido al caracter secuencial del esquema de programación empleado dichas cantidades, no tienen por que coincidir para un
        instante temporal arbitrario. En consecuencia resulta necesario diferenciarlas para que no ocurran interferencias ni desajustes
        (sin ir más lejos, el valor del pulso de conductancia g_out depende fuertemente del paso de la simulación, por tanto, una
        neurona puede haber avanzado de simulation step pero quedarle aún neuronas postsinápticas a las que influenciar que formalmente
        se encuentren en el simulation step anterior, de esta forma, la identificación ciega de los conceptos expuestos resultaría
        catastrófica)

        """

     # (1) CONSIDERACIONES INICIALES

        if self.has_been_simulated:
            self.simulation_step    = 0
            self.has_been_simulated = False
            self.V_membrane         = self.V_initial

        if self.simulation_step == 0:
            self.dt                        = dt
            self.emmited_conductance_pulse = self.emmited_conductance_pulse_calc()

        self.emmited = False
        V_previous   = deepcopy(self.V_membrane)


     # (2) CRIBADO DE LAST_SPIKES

        for target_neuron_id in self.general_neuron_ides:
            for target_receptor_id in self.postsynaptic_receptors_dic[target_neuron_id]["IONTROPIC"].keys():

                dic =  self.general_receptors_dic[target_neuron_id]["IONTROPIC"][self.excitability][target_receptor_id]

                conduction_delay       = dic["conduction_delay"]
                tau_d                  = dic["tau_d"]

                steps_conduction_delay = int(round(conduction_delay/self.dt))
                steps_for_dismishing   = int(round(6*tau_d/self.dt))

                last_spikes            = self.last_spikes[target_neuron_id][target_receptor_id]
                if (len(last_spikes) > 0 and simulation_step - last_spikes[0] >= steps_for_dismishing + steps_conduction_delay):
                    self.last_spikes[target_neuron_id][target_receptor_id].pop(0)


     # (3) AVANCE POR RUNGE-KUTTA 2

        k1 = self.differential_equation_function(simulation_step = simulation_step,     V = self.V_membrane          )
        k2 = self.differential_equation_function(simulation_step = simulation_step + 1, V = self.V_membrane + self.dt * k1)
        self.V_membrane += self.dt*(k1 + k2)/2.0


     # (4) CRIBADO DE NEURONAS EN REFRACTARIO

        steps_refractory      = round(self.time_refractory/self.dt)
        arbitrary_neuron_id   = self.neuron_id
        arbitrary_receptor_id = list(self.postsynaptic_receptors_dic[arbitrary_neuron_id]["IONTROPIC"].keys())[0]
        last_spikes           = self.last_spikes[arbitrary_neuron_id][arbitrary_receptor_id]

        if(len(last_spikes) > 0 and (simulation_step - last_spikes[-1] <= steps_refractory)):
            self.V_membrane = V_previous


     # (5) ACTUALIZACIÓN DE LA INFORMACIÓN

        if (self.V_membrane >= self.V_threshold):
            for target_neuron_id in self.general_neuron_ides:
                for target_receptor_id in self.postsynaptic_receptors_dic[target_neuron_id]["IONTROPIC"].keys():
                    self.last_spikes[target_neuron_id][target_receptor_id].append(simulation_step)

            emmision_time = self.dt * simulation_step + self.dt * (self.V_threshold - V_previous) / (self.V_membrane-V_previous)
            self.emmision_times.append(emmision_time)

            numerator     = (self.V_membrane - self.V_threshold) * (1 + self.dt * (V_previous - self.V_rest))
            denominator   = (self.tau_membrane) * (self.V_membrane - V_previous)
            self.V_membrane = self.V_rest + numerator/denominator

            self.emmited    = True

        self.simulation_step += 1


    def __repr__ (self):

        """
        Representación formal de la neurona en la que se plasma la información identificativa básica.

        """
        return("General LIF Neuron")


    def __str__ (self):

        """
        Representación en forma de cadena de caracteres de la neurona, en la que se detalla toda su información relevante.

        Una eventual adición de receptores metabotrópicos al esquema de programación requerirá del tratamiento exhaustivo de la cadena
        str_metabotropic_receptors a fin de que esta recoja adecuadamente los tipos de receptores de esta clase así como sus propiedades
        más remarcables.

        """

        str_title = "-------------------------------------\
        \n{0}\n-------------------------------------\
        ".format(self.title_str)

        str_LIF   = "\n \
        \n    (1) LIF MODEL DYNAMICS\
        \n        ------------------\
        \n\tThis is a Leaky Integrate and Fire (LIF) neuron. The LIF passive dynamics are governned by the ODE \
        \n\t                                       C(dVm/dt) = Irest(t) + Isyn(t) + Iext\
        \n\twhere\
        \n\t\t· Vm                            : the membrane potential\
        \n\t\t· Iext                          : the injected intensity\
         (current due to an external source)\
        \n\t\t· Irest(t) = -gL[V(t)-Vrest]    : the resting intensity\
          (current through ion channels)\
        \n\t\t· Isyn(t) = -gsyn(t)[V(t)-Vsyn] : the synaptic intensity\
         (current due to a synaptic connection)\
        \n\twith gsyn being the synaptic conductance pulse \
        \n\t                                      gsyn(t) = g{exp[-t/tau_d]-exp[-t/tau_r]} \
        \n\tdefined by\
        \n\t\t· g     : the synaptic coupling strength\
        \n\t\t· tau_d : the decay time of the pulse\
        \n\t\t· tau_r : the rise time of the pulse"

        str_parameters = "\n \
        \n    (2) NEURON PARAMETERS\
        \n        ------------------\
        \n\tThis actual LIF {0} has the following LIF parameters:\
        \n\n\t\t(1) Membrane time constant (tau = C/gL)        : {1:9.3f} ms\
        \n\t\t(2) Time in refractory period                  : {2:9.3f} ms\
        \
        \n\n\t\t(3) Resting membrane potential (Vrest)         : {3:9.3f} mV\
        \n\t\t(4) Membrane potential threshold               : {4:9.3f} mV\
        \
        \n\n\t\t(5) Membrane conductance (gL)                  : {5:9.3f} nS\
        \
        \n\n\t\t(6) External electrode intensity (Iext)        : {6:9.3f} pA,      randomly selected in the range {7} pA\
        \n\t\t(7) Initial membrane potential                 : {8:9.3f} mV,      randomly selected in the range {9} mV\
        ".format(self.neuron_type_str, self.tau_membrane, self.time_refractory,
         self.V_rest, self.V_threshold, self.g_membrane, self.I_ext,
         self.I_ext_range, self.V_membrane_initial, self.V_membrane_initial_range)

        str_iontropic_receptors = "\n \
        \n    (3) IONTROPIC SYNAPTIC RECEPTORS \
        \n        ----------------------------\
        \n\tThe neuron possesses the following iontropic synaptic receptors:"

        for excitability in self.self_receptors_dic["IONTROPIC"].keys():
            str_iontropic_receptors = str_iontropic_receptors + "\
            \n\n\t\t - {0}".format(excitability)
            for receptor in self.self_receptors_dic["IONTROPIC"][excitability].keys():

                str_iontropic_receptors += "\
                \n\n\t\t\t{0}\
                \n\t\t\t\t· Synaptic potential (Vsyn)      : {1:9.3f} mV \
                \n\t\t\t\t· Synaptic coupling strength (g) : {2:9.3f} nS \
                \n\t\t\t\t· Conduction delay               : {3:9.3f} ms \
                \n\t\t\t\t· Rise time (tau_r)              : {4:9.3f} ms \
                \n\t\t\t\t· Decay time (tau_d)             : {5:9.3f} ms \
                ".format(receptor,
                self.self_receptors_dic["IONTROPIC"][excitability][receptor]["V_syn"],
                self.self_receptors_dic["IONTROPIC"][excitability][receptor]["coupling_strength"],
                self.self_receptors_dic["IONTROPIC"][excitability][receptor]["conduction_delay"],
                self.self_receptors_dic["IONTROPIC"][excitability][receptor]["tau_r"],
                self.self_receptors_dic["IONTROPIC"][excitability][receptor]["tau_d"])

        str_metabotropic_receptors = ""

        if self.self_receptors_dic["METABOTROPIC"] != {}:
            str_metabotropic_receptors = "\n \
            \n    (4) METABOTROPIC SYNAPTIC RECEPTORS \
            \n        -------------------------------\
            \n\tThe neuron possesses the following metabotropic synaptic receptors:"

            for receptor in self.self_receptors_dic["METABOTROPIC"].keys():

                string_metabotropic_receptors = string_metabotropic_receptors + "\
                \n\n\t\t\t{0}\
                \n\t\t\t\t· ¡aqui debe ir la descripción de la propiedad 1!\
                \n\t\t\t\t· ¡aqui debe ir la descripción de la propiedad 2!\
                \n\t\t\t\t· etc. : \
                ".format()


        return(str_title + str_LIF + str_parameters + str_iontropic_receptors + str_metabotropic_receptors)


#--------------------------------------------------------------------------------------------------------------------------------------
# CLASES CELULARES DE CUARTO ORDEN (CUARTA GENERACIÓN)
#--------------------------------------------------------------------------------------------------------------------------------------

class Piramidal_NC (LIFN):

    """
    Neurona piramidal genérica del neocortex humano, inherentemente glutamatérgica (por tanto, excitatoria) y proyectiva. Modelizada en
    el marco LIF y con parámetros por defecto usuales [1]

    EJEMPLO
    -------

    Pueden modificarse parámetros
      P = Piramidal_NC()
      L.V_membrane_range = (-75.0,-57.0)

    REFERENCIAS
    -----------
    Durante la redacción se han hecho referencias a las siguientes publicaciones
       [1] N. Brunel and X.-J. Wang, Journal of Neurophysiology 90, 415 (2003)
    """

    neuron_id       = "P_NC"
    excitability    = "EXCITATORY"
    title_str       = "LIF NEOCORTICAL PIRAMIDAL NEURON"
    neuron_type_str = "neocortical piramidal neuron, inherently glutamatergic (excitatory) and proyective"


    def __init__ (self, general_parameter_dic = LIFN.default_parameter_dic, general_receptor_dic = LIFN.default_receptors_dic):

        """
        Estructura de inicialización de la metaclase Piramidal_NC.

        ARGUMENTOS
        ----------

        La inicialización de la metaclase Piramidal_NC admite dos argumentos opcionales
          - general_parameter_dic (dic): diccionario que contiene los parámetros generales de todos los tipos neuronales considerados
          - general_receptor_dic  (dic): diccionario que contiene los parámetros generales de todos los receptores considerados
        Los valores por defecto para los argumentos son los valores por defecto especificados en la metaclase LIFN.

        ATRIBUTOS DE INSTANCIA
        -----------------------

        Toda la información relevante a cerca de los atributos de instancia queda recogida en las anotaciones generales de la metaclase
        LIFN, cuya estructura de inicialización es heredada por la clase Piramidal_NC.

        """

        LIFN.__init__ (self, general_parameter_dic, general_receptor_dic)

    def __repr__ (self):

        """
        Representación formal de la neurona en la que se plasma la información identificativa básica.

        """

        return("PIRAMIDAL NEURON ({}) (LIF, Neocortex)".format(self.neuron_num))




class Interneuron_NC (LIFN):

    """
    Interneurona genérica del neocortex humano, inherentemente GABAérgica (por tanto, inhibitoria) e intrínseca. Modelizada en el marco
    LIF y con parámetros por defecto usuales [1]

    EJEMPLO
    -------

    Pueden modificarse parámetros
      I = Interneuron_NC()
      I.V_membrane_range = (-75.0,-57.0)

    REFERENCIAS
    -----------

    Durante la redacción se han hecho referencia a las siguientes publicaciones
       [1] N. Brunel and X.-J. Wang, Journal of Neurophysiology 90, 415 (2003).

    """

    neuron_id       = "I_NC"
    excitability    = "INHIBITORY"
    title_str       = "LIF NEOCORTICAL INTERNEURON"
    neuron_type_str = "neocortical interneuron, inherently GABAergic (inhibitory) and intrinsic"


    def __init__ (self, general_parameter_dic = LIFN.default_parameter_dic, general_receptor_dic  = LIFN.default_receptors_dic):

        """
        Estructura de inicialización de la metaclase Interneuron_NC.

        ARGUMENTOS
        ----------

        La inicialización de la metaclase Interneuron_NC admite dos argumentos opcionales
          - general_parameter_dic (dic): diccionario que contiene los parámetros generales de todos los tipos neuronales considerados
          - general_receptor_dic  (dic): diccionario que contiene los parámetros generales de todos los receptores considerados
        Los valores por defecto para los argumentos son los valores por defecto especificados en la metaclase LIFN.

        ATRIBUTOS DE INSTANCIA
        -----------------------

        Toda la información relevante a cerca de los atributos de instancia queda recogida en las anotaciones generales de la metaclase
        LIFN, cuya estructura de inicialización es heredada por la clase Interneuron_NC.

        """

        LIFN.__init__ (self, general_parameter_dic, general_receptor_dic)

    def __repr__ (self):

        """
        Representación formal de la neurona en la que se plasma la información identificativa básica.

        """

        return("INTERNEURON ({}) (LIF, Neocortex)".format(self.neuron_num))

#--------------------------------------------------------------------------------------------------------------------------------------
# PRUEBAS
#--------------------------------------------------------------------------------------------------------------------------------------

if __name__ == "__main__":

    import matplotlib.pyplot as plt
    import seaborn           as sns

    P = Interneuron_NC()

    tau_d = P.self_receptors_dic["IONTROPIC"]["INHIBITORY"]["GABA"]["tau_d"]
    tau_r = P.self_receptors_dic["IONTROPIC"]["INHIBITORY"]["GABA"]["tau_r"]
    g     = P.self_receptors_dic["IONTROPIC"]["INHIBITORY"]["GABA"]["coupling_strength"]

    t     = np.arange(0.0,6*tau_d, 0.05)
    y     = P.conductance_pulse_function(t,g,tau_d,tau_r)

    sns.set()
    plt.plot(t, y)
    plt.xlabel("Time (ms)", style = "italic")
    plt.ylabel(r"$g_{syn}$ (nS)", style = "italic")
    plt.ylim(-0.4, 8.0)
    plt.xlim(-1, 30.)
    plt.show()



    # I = Interneuron_NC
    # print(I)
    #
    # I.add_receptor(receptor_type = "IONTROPIC", excitability = "EXCITATORY",
    #                receptor_name = "NDMA", Vsyn = -20.0,
    #                coupling_strength = 0.3, conduction_delay = 0.0,
    #                tau_r = 0.3, tau_d = 2)
    # print(I)
    #
    # I.delete_receptor(receptor_type = "IONTROPIC",
    #                   excitability = "EXCITATORY", receptor_name = "AMPA")
    # print(I)
    #
    # print(I.receptors_dic)
    # print(I.postsynaptic_receptors_dic)
