"""
---------------------------------------------------------------------------------------------------------------------------------------
MÓDULO COEFFICIENTS
---------------------------------------------------------------------------------------------------------------------------------------

El módulo coefficients contiene un programa autoexplicativo durante la ejecución que permite representaciones complejas y muy variadas
sobre las diferentes métricas que aspiran a cuantificar el fenómeno SW.

"""

from wattsstrogatz         import *
from connectome            import *
from smallworldiness       import *
from collections           import defaultdict
from pandas                import Series, DataFrame
from termcolor             import colored
from math                  import ceil
from datetime              import datetime

import pandas              as pd
import matplotlib.pyplot   as plt
import matplotlib.gridspec as gridspec
import os                  as os
import seaborn             as sns
import numpy               as np

#--------------------------------------------------------------------------------------------------------------------------------------
# GENERAL FUNCTIONS
#--------------------------------------------------------------------------------------------------------------------------------------

def drange(x, y, step, precission = None):

    """
    Extiende la función intrínseca range a intervalos decimales aspirando aminimizar el efecto de la precisión limitada.

    ARGUMENTOS
    ----------
    x         (float) : número inicial de la secuencia (incluido)
    y         (float) : número final de la secuencia (excluido)
    step      (float) : salto de un elemento a otro en la secuencia
    precision (int)   : número de decimales de precisión
    """
    while round(x,precission) < round(y, precission):
        if precission == None:
            yield float(x)
        else:
            yield round(x,precission)
        x += step

def default_to_regular(d):
    """
    Convierte un arbol de default_dict anidados a usual dict, si d no es un defaultdict devuelve d inmutado.

    """
    if isinstance(d, defaultdict):
        d = {k: default_to_regular(v) for k, v in d.items()}
    return d

#--------------------------------------------------------------------------------------------------------------------------------------
# DESARROLLO
#--------------------------------------------------------------------------------------------------------------------------------------

if __name__ == "__main__":

    start_time = datetime.now()

    i = 0
    j = 0
    repeat = True
    while repeat:
     # (1) LECTURA DE DATOS
        print("--------------------------------------------------")
        print("SMALL WORLD COEFFICIENTS (SIGMA, OMEGA, SWI)")
        print("--------------------------------------------------")
        print("""Hello! This is a program for obtaining Small World coefficient values. Next you are going to be asked to provide values for some parameters, keep in mind that:
           (1) Values ought to be separated by colonns if more than one provided.
           (2) Values of q must be even and values of p must lay in the interval [0.0, 1.0].
           (3) Variations will be considered in every parameter you provide more than a single value for.
           (4) Equiespaciated intervals are allowed (bear in mind that 'start' is included, 'end' excluded). If such an interval is preffered for any of the parameters write any string when you are asked for its value. """)
        print(colored("IMPORTANT!","red")+" Make sure EXECUTION directorie DOES NOT exist in your CWD, if it does, rename or move it.")

        graph_type = input("\nIntroduce the type of graph you want to consider (directed or undirected): ")

        try:
            N = [int(x) for x in input("\nIntroduce the number(s) of nodes                         : ").split(",")]
        except ValueError:
            N_range = [int(x) for x in input("\tIntroduce in order: start, end, step: ").split(",")]
            N       = list(range(N_range[0], N_range[1], N_range[2]))

        try:
            q = [int(x) for x in input("Introduce the even number(s) of initial nearest neighbors: ").split(",")]

        except ValueError:
            q_range = [int(x) for x in input("\tIntroduce in order: start, end, step: ").split(",")]
            q       = list(range(q_range[0], q_range[1], q_range[2]))

        try:
            p = [float(x) for x in input("Introduce the initial probability/probabilities          : ").split(",")]

        except ValueError:
            p_range = [float(x) for x in input("\tIntroduce in order: start, end, step, precision: ").split(",")]
            p       = list(drange(p_range[0], p_range[1], p_range[2], int(p_range[3])))

        try:
            nrand = [int(x) for x in input("\nIntroduce nrand, the number of randomized graphs that will be created : ").split(",")]
        except:
            nrand_range = [int(x) for x in input ("\tIntroduce in order: start, end, step: ").split(",")]
            nrand       = list(range(nrand_range[0], nrand_range[1], nrand_range[2]))

        try:
            nlat = [int(x) for x in input("Introduce nlat, the number of latticized graphs that will be created  : ").split(",")]
        except:
            nlat_range = [int(x) for x in input ("\tIntroduce in order: start, end, step: ").split(",").split(",")]
            nlat       = list(range(nlat_range[0], nlat_range[1], nlat_range[2]))

        try:
            niter = [int(x) for x in input("Introduce niter, the number of aproximate times a swapp will occur    : ").split(",")]
        except:
            niter_range = [int(x) for x in input ("\tIntroduce in order: start, end, step: ").split(",")]
            niter       = list(range(niter_range[0], niter_range[1], niter_range[2]))


        print("\n----------------------------------------------------------------------------------------------------------------------------------------------------")
        print("The values considered for N are   : {}".format(str(N)[1:-1]))
        print("The values considered for q are   : {}".format(str(q)[1:-1]))
        print("The values considered for p are   : {}".format(str(p)[1:-1]))
        print("The values considered for nrand is: {}".format(str(nrand)[1:-1]))
        print("The values considered for nlat  is: {}".format(str(nlat)[1:-1]))
        print("The values considered for niter is: {}".format(str(niter)[1:-1]))
        print("----------------------------------------------------------------------------------------------------------------------------------------------------")

        ntotal = len(N)*len(q)*len(p)*len(nrand)*len(nlat)*len(niter)
        npart  = len(N)*len(q)*len(nrand)*len(nlat)*len(niter)

        print("\nThe program is capable of:")
        graph_files = input("\t(1) Creating GraphML files for every SmallWorld, Randomized and Latticized graph created, do you want the program to do so? ({} files in total): ".format(ntotal))
        data_files  = input("\t(2) Creating CSV files with the coefficients data, do you want the program to do so? ({} files in total)                                        : ".format(npart))
        data_plot   = input("\t(3) Creating PDF format plots with the coefficients data, do you want the program to do so? ({} plots in total)                                 : ".format(npart))

     # (2) OBTENCIÓN DE RESULTADOS
     # La información obtenida (coeficientes de Small Worldiness los grafos y medidas ausiliares utilizadas para su determinación) se organiza en un único diccionario anidado (nested)
     # de forma que toda la información resulte fácilmente accesible e intuitivamente navegable.

        print("\nThe program is about to start its execution!")

        recursive_dictionary = lambda : defaultdict(recursive_dictionary) # permite inicializar un diccionario sobre la marcha
        WS_results           = recursive_dictionary()


        for n_niter in niter:
            print("Starting analysis of niter = {}".format(n_niter))
            for n_nrand in nrand:
                print("\tStarting analysis of nrand = {}".format(n_nrand))
                for n_nlat in nlat:
                    print("\t\tStarting analysis of nlat = {}".format(n_nlat))
                    for nnodes in N:
                        print("\t\t\tStarting analysis of N = {}".format(nnodes))
                        for neighbors in q:
                            print("\t\t\t\tStarting analysis of q = {}".format(neighbors))
                            for probability in p:
                                print("\t\t\t\t\tStarting analysis of p = {}".format(probability))

                                if graph_type.lower()[0] == "u":
                                    WS = WattsStrogatz(N = nnodes, q = neighbors, p = probability)
                                elif graph_type.lower()[0] == "d":
                                    WS = DiWattsStrogatz(N = nnodes, q = neighbors, p = probability)

                                randomized = [randomize(WS, niter = n_niter) for i in range(n_nrand)]
                                latticized = [latticize(WS, niter = n_niter) for i in range(n_nlat )]

                                WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["SW COEFFICIENTS"]["Sigma"]          = sigma  (WS, RG = randomized)
                                WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["SW COEFFICIENTS"]["Omega_p"]        = omega_p(WS, RG = randomized, LG = latticized)
                                WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["SW COEFFICIENTS"]["SWI"], coef_dict = SWI    (WS, RG = randomized, LG = latticized, sup = True)

                                dict_int = {"C":[nx.average_clustering(WS)],"L":[nx.average_shortest_path_length(WS)]}

                                WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["AUXILIAR MEASUREMENTS"]          = {**coef_dict,**dict_int}
                                WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["GRAPHS"]["WattsStrogatz"]        = WS
                                WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["GRAPHS"]["Randomized"]           = randomized
                                WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["GRAPHS"]["Latticized"]           = latticized


        WS_results  = default_to_regular(WS_results)

     # (3) CUSTOMIZACIÓN DE RESPUESTA

        print("\n")

        unsolved = True
        while unsolved:
            try:
                path_origen = os.getcwd()+"/EXECUTION"
                os.mkdir(path_origen)
                unsolved = False
            except FileExistsError:
                unsolved = True
                input(colored("A FILE NAMED 'EXECUTION' ALREADY EXISTS IN YOUR CWD, please, delete or move it. Write OK when solved: ","red"))

        print("\n")


        if graph_files.lower()[0] == "y":
            path_g = os.getcwd()+"/EXECUTION/GRAPHS(GraphML)"
            os.mkdir(path_g)
            for n_niter in niter:
                for n_nrand in nrand:
                    for n_nlat in nlat:
                        for nnodes in N:
                            for neighbors in q:
                                for probability in p:
                                    nx.write_graphml(WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["GRAPHS"]["WattsStrogatz"],path = path_g+"/Gsmallworld(N = {}, q = {}, p = {}, nrand = {}, niter = {}).graphml".format(nnodes, neighbors, probability, nrand, niter))
                                    for G in WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["GRAPHS"]["Randomized"]:
                                        nx.write_graphml(G, path = path_g+"/Granodmized(N = {}, q = {}, p = {}, nrand = {}, niter = {}).graphml".format(nnodes, neighbors, probability, n_nrand, n_niter))
                                    for G in WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][probability]["GRAPHS"]["Latticized"]:
                                        nx.write_graphml(G, path = path_g+"/Glatticized(N = {}, q = {}, p = {}, nrand = {}, niter = {}).graphml".format(nnodes, neighbors, probability, n_nrand, n_niter))
            print("......SUCCESS! The GraphML files have been succesfully created in your working directory.")

        if data_files.lower()[0] == "y" or data_plot.lower()[0] == "y":

            for n_niter in niter:
                for n_nrand in nrand:
                    for n_nlat in nlat:
                        for nnodes in N:
                            for neighbors in q:
                                data      = DataFrame(index = list(range(len(p))),columns = ["Probabilities", "Sigma", "Omega","SWI", "AuxiliarMeasurements"])
                                data.name = "Small Worldiness, N = {}, q = {}, nrand = {}, nlat = {}, niter = {}".format(nnodes, neighbors, n_nrand, n_nlat, n_niter)

                                data.Probabilities        = p
                                data.Sigma                = [WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][i]["SW COEFFICIENTS"]["Sigma"]   for i in p]
                                data.Omega                = [WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][i]["SW COEFFICIENTS"]["Omega_p"] for i in p]
                                data.SWI                  = [WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][i]["SW COEFFICIENTS"]["SWI"]     for i in p]
                                data.AuxiliarMeasurements = [WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][i]["AUXILIAR MEASUREMENTS"]      for i in p]

                                list_coef_c = []
                                list_coef_l = []
                                for k in p:
                                    dict_interm = WS_results[nnodes][neighbors][n_niter][n_nrand][n_nlat][k]["AUXILIAR MEASUREMENTS"]
                                    for l in dict_interm.keys():
                                        if l[0] == "C":
                                            list_coef_c.append(np.mean(dict_interm[l]))
                                        elif l[0] == "L":
                                            list_coef_l.append(np.mean(dict_interm[l]))
                                max_coef_c = round(max(list_coef_c),1)
                                max_coef_l = round(max(list_coef_l),1)


                                if data_files.lower()[0] == "y":
                                    path_d = os.getcwd()+"/EXECUTION/DataFrames(CSV)"
                                    os.mkdir(path_d)

                                    data.to_csv(path_or_buf = path_d+"/DataFrame(N = {}, q = {}, nrand = {}, nlat = {}, niter = {}).csv".format(nnodes, neighbors, n_nrand, n_nlat, n_niter))
                                    print("......SUCCESS! The CSV files have been succesfully created in your working directory.")

                                if data_plot.lower()[0] == "y":
                                    path_p = os.getcwd()+"/EXECUTION/PLOTS(PDF)"
                                    os.mkdir(path_p)

                                    i += 2
                                    sns.set()
                                    fig = plt.figure(i)
                                    gs  = gridspec.GridSpec(2,2)

                                    ax1 = fig.add_subplot(gs[0,:]) #sigma
                                    ax2 = fig.add_subplot(gs[1:,0]) #omega
                                    ax3 = fig.add_subplot(gs[1:,1:]) #SWI

                                    #ax1.scatter(data.Probabilities, data.Sigma)
                                    ax1.plot(data.Probabilities, data.Sigma)
                                    ax1.set_xlabel("Rewiring Probability (p)", style = "italic", size = 9)
                                    ax1.set_ylabel(r"SW Coefficient ($\sigma$)", style = "italic", size = 9)
                                    ax1.set_title (r'$\sigma$ coefficient', style = "italic", fontweight = "bold", size = 11)

                                    #ax2.scatter(data.Probabilities, data.Omega)
                                    ax2.plot(data.Probabilities, data.Omega)
                                    ax2.set_xlabel("Rewiring Probability (p)", style = "italic", size = 9)
                                    ax2.set_ylabel(r"SW Coefficient ($\omega$')", style = "italic", size = 9)
                                    ax2.set_ylim(-0.1,1.1)
                                    ax2.set_title (r'$\omega$ coefficient', style = "italic", fontweight = "bold", size = 11)

                                    #ax3.scatter(data.Probabilities, data.SWI)
                                    ax3.plot(data.Probabilities, data.SWI)
                                    ax3.set_xlabel("Rewiring Probability (p)", style = "italic", size = 9)
                                    ax3.set_ylabel("SW Coefficient (SWI)", style = "italic", size = 9)
                                    ax3.set_ylim(-0.1,1.1)
                                    ax3.set_title ('SWI coefficient', style = "italic", fontweight = "bold", size = 11)

                                    fig.suptitle(r"$\mathit{N}$ = %d, $\mathit{q}$ = %d, $n_{rand}$ = %d, $n_{lat}$ = %d, $n_{iter}$ = %d" % (nnodes, neighbors, n_nrand, n_nlat, n_niter), size = 12)

                                    gs.tight_layout(fig, rect=[0, 0.03, 1, 0.95])

                                    plt.savefig(fname = path_p+"/SW_coefficients(N = {}, q = {}, n = {}, nlat = {}, niter = {}).pdf".format(nnodes, neighbors, n_nrand, n_nlat, n_niter))


                                    j += 3
                                    sns.set()
                                    fig_h = plt.figure(j, figsize = (20,20))
                                    nrow  = ceil(len(p)/2)
                                    ncol  = 2
                                    gs    = gridspec.GridSpec(nrow,ncol)

                                    ax_c = [fig_h.add_subplot(gs[i,j]) for i in range(nrow) for j in range(ncol)]

                                    for i in range(len(p)):
                                        AUXdict = data.ix[i,"AuxiliarMeasurements"]

                                        heights_c = [np.mean(AUXdict["C"]), np.mean(AUXdict["Cran"]), np.mean(AUXdict["Clat"]), 0.0, 0.0, 0.0]
                                        heights_l = [0.0, 0.0, 0.0, np.mean(AUXdict["L"]), np.mean(AUXdict["Lran"]), np.mean(AUXdict["Llat"])]
                                        centers = [0,1,2,3,4,5]

                                        pc1, pc2, pc3, pc4, pc5, pc6 = ax_c[i].bar(centers,heights_c,align = "center", tick_label = ["C", r"$C_{ran}$", r"$C_{lat}$", "L", r"$L_{ran}$", r"$L_{lat}$"])
                                        pc1.set_facecolor("r")
                                        pc2.set_facecolor("r")
                                        pc3.set_facecolor("r")

                                        ax_l = ax_c[i].twinx()
                                        pl1, pl2, pl3, pl4, pl5, pl6 = ax_l.bar(centers,heights_l,align = "center")

                                        ax_c[i].set_title("p = {}".format(p[i]))

                                        ax_c[i].tick_params(axis="y", labelcolor = "r", color = "r", grid_alpha = 0.0, length = 4, width = 0.5, labelsize = 8)
                                        ax_l.tick_params(axis="y", labelcolor ="b", color = "b", grid_alpha = 0.0, length = 4, width = 0.5, labelsize = 8)

                                        ax_c[i].set_ylim(0,max_coef_c)
                                        ax_l.set_ylim(0,max_coef_l)

                                    if len(p) % 2 != 0:
                                        ax_c[-1].axis("off")

                                    fig_h.suptitle(r"$\mathit{N}$ = %d, $\mathit{q}$ = %d, $n_{rand}$ = %d, $n_{lat}$ = %d, $n_{iter}$ = %d" % (nnodes, neighbors, n_nrand, n_nlat, n_niter), size = 12)

                                    gs.tight_layout(fig_h, rect = [0,0.03, 1, 0.95])

                                    plt.savefig(fname = path_p+"/BARS(N = {}, q = {}, n = {}, nlat = {}, niter = {}).pdf".format(nnodes, neighbors, n_nrand, n_nlat, n_niter))

                                    print("......SUCCESS! The PDF files have been succesfully created in your working directory.")

        path_readme  = os.getcwd()+"/EXECUTION"
        filename     = path_readme+"/README.txt"

        readme = open(filename, "w")
        readme.write("In this directory some subdirectories can be find containing different types of information provided by the program coefficients.py")
        readme.write("\n\tGraph type: {}".format(graph_type.upper()))
        end_time = datetime.now()
        readme.write("\n\tExecution time: {}".format(end_time-start_time))
        readme.write("\n\n\tN     = {}\n\tq     = {}\n\tp     = {}\n\tnrand = {}\n\tnlat  = {}\n\tniter = {}".format(str(N)[1:-1], str(q)[1:-1], str(p)[1:-1], str(nrand)[1:-1], str(nlat)[1:-1], str(niter)[1:-1]))

        readme.close()


     # (4) POSIBILIDAD DE REEJECUCIÓN
        another = input("\nThe program has come to its end, thank you! Want another execution?: ")
        if another.lower()[0] == "y":
            repeat = True
        else:
            repeat = False

        print("\n")
